import pytest
import sqlalchemy as sa

from sqlalchemy_utils import (
    create_materialized_view,
    create_view,
    refresh_materialized_view
)


@pytest.fixture
def Article(Base, User):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
        author_id = sa.Column(sa.Integer, sa.ForeignKey(User.id))
        author = sa.orm.relationship(User)
    return Article


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
    return User


@pytest.fixture
def ArticleMV(Base, Article, User):
    class ArticleMV(Base):
        __table__ = create_materialized_view(
            name='article_mv',
            selectable=sa.select(
                [
                    Article.id,
                    Article.name,
                    User.id.label('author_id'),
                    User.name.label('author_name')
                ],
                from_obj=(
                    Article.__table__
                    .join(User, Article.author_id == User.id)
                )
            ),
            aliases={'name': 'article_name'},
            metadata=Base.metadata,
            indexes=[sa.Index('article_mv_id_idx', 'id')]
        )
    return ArticleMV


@pytest.fixture
def ArticleView(Base, Article, User):
    class ArticleView(Base):
        __table__ = create_view(
            name='article_view',
            selectable=sa.select(
                [
                    Article.id,
                    Article.name,
                    User.id.label('author_id'),
                    User.name.label('author_name')
                ],
                from_obj=(
                    Article.__table__
                    .join(User, Article.author_id == User.id)
                )
            ),
            metadata=Base.metadata
        )
    return ArticleView


@pytest.fixture
def init_models(ArticleMV, ArticleView):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestMaterializedViews:
    def test_refresh_materialized_view(
        self,
        session,
        Article,
        User,
        ArticleMV
    ):
        article = Article(
            name='Some article',
            author=User(name='Some user')
        )
        session.add(article)
        session.commit()
        refresh_materialized_view(session, 'article_mv')
        materialized = session.query(ArticleMV).first()
        assert materialized.article_name == 'Some article'
        assert materialized.author_name == 'Some user'

    def test_querying_view(
        self,
        session,
        Article,
        User,
        ArticleView
    ):
        article = Article(
            name='Some article',
            author=User(name='Some user')
        )
        session.add(article)
        session.commit()
        row = session.query(ArticleView).first()
        assert row.name == 'Some article'
        assert row.author_name == 'Some user'


class TrivialViewTestCases:
    def life_cycle(
        self,
        engine,
        metadata,
        column,
        cascade_on_drop
    ):
        __table__ = create_view(
            name='trivial_view',
            selectable=sa.select([column]),
            metadata=metadata,
            cascade_on_drop=cascade_on_drop
        )
        __table__.create(engine)
        __table__.drop(engine)


class SupportsCascade(TrivialViewTestCases):
    def test_life_cycle_cascade(
        self,
        connection,
        engine,
        Base,
        User
    ):
        self.life_cycle(engine, Base.metadata, User.id, cascade_on_drop=True)


class DoesntSupportCascade(SupportsCascade):
    @pytest.mark.xfail
    def test_life_cycle_cascade(self, *args, **kwargs):
        super(DoesntSupportCascade, self).test_life_cycle_cascade(
            *args,
            **kwargs
        )


class SupportsNoCascade(TrivialViewTestCases):
    def test_life_cycle_no_cascade(
        self,
        connection,
        engine,
        Base,
        User
    ):
        self.life_cycle(engine, Base.metadata, User.id, cascade_on_drop=False)


@pytest.mark.usefixtures('postgresql_dsn')
class TestPostgresTrivialView(SupportsCascade, SupportsNoCascade):
    pass


@pytest.mark.usefixtures('mysql_dsn')
class TestMySqlTrivialView(SupportsCascade, SupportsNoCascade):
    pass


@pytest.mark.usefixtures('sqlite_none_database_dsn')
class TestSqliteTrivialView(DoesntSupportCascade, SupportsNoCascade):
    pass



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.observer import observes


@pytest.fixture
def Catalog(Base):
    catalog_category = sa.Table(
        'catalog_category',
        Base.metadata,
        sa.Column('catalog_id', sa.Integer, sa.ForeignKey('catalog.id')),
        sa.Column('category_id', sa.Integer, sa.ForeignKey('category.id'))
    )

    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        product_count = sa.Column(sa.Integer, default=0)

        @observes('categories.sub_categories.products')
        def product_observer(self, products):
            self.product_count = len(products)

        categories = sa.orm.relationship(
            'Category',
            backref='catalogs',
            secondary=catalog_category
        )
    return Catalog


@pytest.fixture
def Category(Base):
    category_subcategory = sa.Table(
        'category_subcategory',
        Base.metadata,
        sa.Column(
            'category_id',
            sa.Integer,
            sa.ForeignKey('category.id')
        ),
        sa.Column(
            'subcategory_id',
            sa.Integer,
            sa.ForeignKey('sub_category.id')
        )
    )

    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)

        sub_categories = sa.orm.relationship(
            'SubCategory',
            backref='categories',
            secondary=category_subcategory
        )
    return Category


@pytest.fixture
def SubCategory(Base):
    subcategory_product = sa.Table(
        'subcategory_product',
        Base.metadata,
        sa.Column(
            'subcategory_id',
            sa.Integer,
            sa.ForeignKey('sub_category.id')
        ),
        sa.Column(
            'product_id',
            sa.Integer,
            sa.ForeignKey('product.id')
        )
    )

    class SubCategory(Base):
        __tablename__ = 'sub_category'
        id = sa.Column(sa.Integer, primary_key=True)
        products = sa.orm.relationship(
            'Product',
            backref='sub_categories',
            secondary=subcategory_product
        )

    return SubCategory


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        price = sa.Column(sa.Numeric)
    return Product


@pytest.fixture
def init_models(Catalog, Category, SubCategory, Product):
    pass


@pytest.fixture
def catalog(session, Catalog, Category, SubCategory, Product):
    sub_category = SubCategory(products=[Product()])
    category = Category(sub_categories=[sub_category])
    catalog = Catalog(categories=[category])
    session.add(catalog)
    session.flush()
    return catalog


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForManyToManyToManyToMany(object):

    def test_simple_insert(self, catalog):
        assert catalog.product_count == 1

    def test_add_leaf_object(self, catalog, session, Product):
        product = Product()
        catalog.categories[0].sub_categories[0].products.append(product)
        session.flush()
        assert catalog.product_count == 2

    def test_remove_leaf_object(self, catalog, session, Product):
        product = Product()
        catalog.categories[0].sub_categories[0].products.append(product)
        session.flush()
        session.delete(product)
        session.flush()
        assert catalog.product_count == 1

    def test_delete_intermediate_object(self, catalog, session):
        session.delete(catalog.categories[0].sub_categories[0])
        session.commit()
        assert catalog.product_count == 0

    def test_gathered_objects_are_distinct(
        self,
        session,
        Catalog,
        Category,
        SubCategory,
        Product
    ):
        catalog = Catalog()
        category = Category(catalogs=[catalog])
        product = Product()
        category.sub_categories.append(
            SubCategory(products=[product])
        )
        session.add(
            SubCategory(categories=[category], products=[product])
        )
        session.commit()
        assert catalog.product_count == 1



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.aggregates import aggregated


@pytest.fixture
def Thread(Base):
    class Thread(Base):
        __tablename__ = 'thread'
        id = sa.Column(sa.Integer, primary_key=True)

        @aggregated(
            'comments',
            sa.Column('_comment_count', sa.Integer, default=0)
        )
        def comment_count(self):
            return sa.func.count('1')

        comments = sa.orm.relationship('Comment', backref='thread')
    return Thread


@pytest.fixture
def Comment(Base):
    class Comment(Base):
        __tablename__ = 'comment'
        id = sa.Column(sa.Integer, primary_key=True)
        thread_id = sa.Column(sa.Integer, sa.ForeignKey('thread.id'))
    return Comment


@pytest.fixture
def init_models(Thread, Comment):
    pass


class TestAggregatedWithColumnAlias(object):

    def test_assigns_aggregates_on_insert(self, session, Thread, Comment):
        thread = Thread()
        session.add(thread)
        comment = Comment(thread=thread)
        session.add(comment)
        session.commit()
        session.refresh(thread)
        assert thread.comment_count == 1

    def test_assigns_aggregates_on_separate_insert(
        self,
        session,
        Thread,
        Comment
    ):
        thread = Thread()
        session.add(thread)
        session.commit()
        comment = Comment(thread=thread)
        session.add(comment)
        session.commit()
        session.refresh(thread)
        assert thread.comment_count == 1

    def test_assigns_aggregates_on_delete(self, session, Thread, Comment):
        thread = Thread()
        session.add(thread)
        session.commit()
        comment = Comment(thread=thread)
        session.add(comment)
        session.commit()
        session.delete(comment)
        session.commit()
        session.refresh(thread)
        assert thread.comment_count == 0



import six


class GenericRelationshipTestCase(object):
    def test_set_as_none(self, Event):
        event = Event()
        event.object = None
        assert event.object is None

    def test_set_manual_and_get(self, session, User, Event):
        user = User()

        session.add(user)
        session.commit()

        event = Event()
        event.object_id = user.id
        event.object_type = six.text_type(type(user).__name__)

        assert event.object is None

        session.add(event)
        session.commit()

        assert event.object == user

    def test_set_and_get(self, session, User, Event):
        user = User()

        session.add(user)
        session.commit()

        event = Event(object=user)

        assert event.object_id == user.id
        assert event.object_type == type(user).__name__

        session.add(event)
        session.commit()

        assert event.object == user

    def test_compare_instance(self, session, User, Event):
        user1 = User()
        user2 = User()

        session.add_all([user1, user2])
        session.commit()

        event = Event(object=user1)

        session.add(event)
        session.commit()

        assert event.object == user1
        assert event.object != user2

    def test_compare_query(self, session, User, Event):
        user1 = User()
        user2 = User()

        session.add_all([user1, user2])
        session.commit()

        event = Event(object=user1)

        session.add(event)
        session.commit()

        q = session.query(Event)
        assert q.filter_by(object=user1).first() is not None
        assert q.filter_by(object=user2).first() is None
        assert q.filter(Event.object == user2).first() is None

    def test_compare_not_query(self, session, User, Event):
        user1 = User()
        user2 = User()

        session.add_all([user1, user2])
        session.commit()

        event = Event(object=user1)

        session.add(event)
        session.commit()

        q = session.query(Event)
        assert q.filter(Event.object != user2).first() is not None

    def test_compare_type(self, session, User, Event):
        user1 = User()
        user2 = User()

        session.add_all([user1, user2])
        session.commit()

        event1 = Event(object=user1)
        event2 = Event(object=user2)

        session.add_all([event1, event2])
        session.commit()

        statement = Event.object.is_type(User)
        q = session.query(Event).filter(statement)
        assert q.first() is not None



# -*- coding: utf-8 -*-
import pytest
import sqlalchemy as sa

from sqlalchemy_utils.types import json


@pytest.fixture
def Document(Base):
    class Document(Base):
        __tablename__ = 'document'
        id = sa.Column(sa.Integer, primary_key=True)
        json = sa.Column(json.JSONType)
    return Document


@pytest.fixture
def init_models(Document):
    pass


class JSONTestCase(object):

    def test_list(self, session, Document):
        document = Document(
            json=[1, 2, 3]
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.json == [1, 2, 3]

    def test_parameter_processing(self, session, Document):
        document = Document(
            json={'something': 12}
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.json == {'something': 12}

    def test_non_ascii_chars(self, session, Document):
        document = Document(
            json={'something': u'äääööö'}
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.json == {'something': u'äääööö'}


@pytest.mark.skipif('json.json is None')
@pytest.mark.usefixtures('sqlite_memory_dsn')
class TestSqliteJSONType(JSONTestCase):
    pass


@pytest.mark.skipif('json.json is None')
@pytest.mark.usefixtures('postgresql_dsn')
class TestPostgresJSONType(JSONTestCase):
    pass



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import merge_references


class TestMergeReferences(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))

            def __repr__(self):
                return 'User(%r)' % self.name
        return User

    @pytest.fixture
    def BlogPost(self, Base, User):
        class BlogPost(Base):
            __tablename__ = 'blog_post'
            id = sa.Column(sa.Integer, primary_key=True)
            title = sa.Column(sa.Unicode(255))
            content = sa.Column(sa.UnicodeText)
            author_id = sa.Column(sa.Integer, sa.ForeignKey('user.id'))

            author = sa.orm.relationship(User)
        return BlogPost

    @pytest.fixture
    def init_models(self, User, BlogPost):
        pass

    def test_updates_foreign_keys(self, session, User, BlogPost):
        john = User(name=u'John')
        jack = User(name=u'Jack')
        post = BlogPost(title=u'Some title', author=john)
        post2 = BlogPost(title=u'Other title', author=jack)
        session.add(john)
        session.add(jack)
        session.add(post)
        session.add(post2)
        session.commit()
        merge_references(john, jack)
        session.commit()
        assert post.author == jack
        assert post2.author == jack

    def test_object_merging_whenever_possible(self, session, User, BlogPost):
        john = User(name=u'John')
        jack = User(name=u'Jack')
        post = BlogPost(title=u'Some title', author=john)
        post2 = BlogPost(title=u'Other title', author=jack)
        session.add(john)
        session.add(jack)
        session.add(post)
        session.add(post2)
        session.commit()
        # Load the author for post
        assert post.author_id == john.id
        merge_references(john, jack)
        assert post.author_id == jack.id
        assert post2.author_id == jack.id


class TestMergeReferencesWithManyToManyAssociations(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))

            def __repr__(self):
                return 'User(%r)' % self.name
        return User

    @pytest.fixture
    def Team(self, Base):
        team_member = sa.Table(
            'team_member', Base.metadata,
            sa.Column(
                'user_id', sa.Integer,
                sa.ForeignKey('user.id', ondelete='CASCADE'),
                primary_key=True
            ),
            sa.Column(
                'team_id', sa.Integer,
                sa.ForeignKey('team.id', ondelete='CASCADE'),
                primary_key=True
            )
        )

        class Team(Base):
            __tablename__ = 'team'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))

            members = sa.orm.relationship(
                'User',
                secondary=team_member,
                backref='teams'
            )
        return Team

    @pytest.fixture
    def init_models(self, User, Team):
        pass

    def test_supports_associations(self, session, User, Team):
        john = User(name=u'John')
        jack = User(name=u'Jack')
        team = Team(name=u'Team')
        team.members.append(john)
        session.add(john)
        session.add(jack)
        session.commit()
        merge_references(john, jack)
        assert john not in team.members
        assert jack in team.members


class TestMergeReferencesWithManyToManyAssociationObjects(object):

    @pytest.fixture
    def Team(self, Base):
        class Team(Base):
            __tablename__ = 'team'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return Team

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return User

    @pytest.fixture
    def TeamMember(self, Base, User, Team):
        class TeamMember(Base):
            __tablename__ = 'team_member'
            user_id = sa.Column(
                sa.Integer,
                sa.ForeignKey(User.id, ondelete='CASCADE'),
                primary_key=True
            )
            team_id = sa.Column(
                sa.Integer,
                sa.ForeignKey(Team.id, ondelete='CASCADE'),
                primary_key=True
            )
            role = sa.Column(sa.Unicode(255))
            team = sa.orm.relationship(
                Team,
                backref=sa.orm.backref(
                    'members',
                    cascade='all, delete-orphan'
                ),
                primaryjoin=team_id == Team.id,
            )
            user = sa.orm.relationship(
                User,
                backref=sa.orm.backref(
                    'memberships',
                    cascade='all, delete-orphan'
                ),
                primaryjoin=user_id == User.id,
            )
        return TeamMember

    @pytest.fixture
    def init_models(self, User, Team, TeamMember):
        pass

    def test_supports_associations(self, session, User, Team, TeamMember):
        john = User(name=u'John')
        jack = User(name=u'Jack')
        team = Team(name=u'Team')
        team.members.append(TeamMember(user=john))
        session.add(john)
        session.add(jack)
        session.add(team)
        session.commit()
        merge_references(john, jack)
        session.commit()
        users = [member.user for member in team.members]
        assert john not in users
        assert jack in users



import pytest
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

from sqlalchemy_utils import generic_relationship

from . import GenericRelationshipTestCase


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column(sa.Integer, primary_key=True)
    return Building


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
    return User


@pytest.fixture
def EventBase(Base):
    class EventBase(Base):
        __abstract__ = True

        object_type = sa.Column(sa.Unicode(255))
        object_id = sa.Column(sa.Integer, nullable=False)

        @declared_attr
        def object(cls):
            return generic_relationship('object_type', 'object_id')
    return EventBase


@pytest.fixture
def Event(EventBase):
    class Event(EventBase):
        __tablename__ = 'event'
        id = sa.Column(sa.Integer, primary_key=True)
    return Event


@pytest.fixture
def init_models(Building, User, Event):
    pass


class TestGenericRelationshipWithAbstractBase(GenericRelationshipTestCase):
    pass



import pytest
import six
import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy_utils import generic_relationship


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
    return User


@pytest.fixture
def UserHistory(Base):
    class UserHistory(Base):
        __tablename__ = 'user_history'
        id = sa.Column(sa.Integer, primary_key=True)

        transaction_id = sa.Column(sa.Integer, primary_key=True)
    return UserHistory


@pytest.fixture
def Event(Base):
    class Event(Base):
        __tablename__ = 'event'
        id = sa.Column(sa.Integer, primary_key=True)

        transaction_id = sa.Column(sa.Integer)

        object_type = sa.Column(sa.Unicode(255))
        object_id = sa.Column(sa.Integer, nullable=False)

        object = generic_relationship(
            object_type, object_id
        )

        @hybrid_property
        def object_version_type(self):
            return self.object_type + 'History'

        @object_version_type.expression
        def object_version_type(cls):
            return sa.func.concat(cls.object_type, 'History')

        object_version = generic_relationship(
            object_version_type, (object_id, transaction_id)
        )
    return Event


@pytest.fixture
def init_models(User, UserHistory, Event):
    pass


class TestGenericRelationship(object):

    def test_set_manual_and_get(self, session, User, UserHistory, Event):
        user = User(id=1)
        history = UserHistory(id=1, transaction_id=1)
        session.add(user)
        session.add(history)
        session.commit()

        event = Event(transaction_id=1)
        event.object_id = user.id
        event.object_type = six.text_type(type(user).__name__)
        assert event.object is None

        session.add(event)
        session.commit()

        assert event.object == user
        assert event.object_version == history



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.types import url


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        website = sa.Column(url.URLType)

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.mark.skipif('url.furl is None')
class TestURLType(object):

    def test_color_parameter_processing(self, session, User):
        user = User(
            website=url.furl(u'www.example.com')
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert isinstance(user.website, url.furl)

    def test_scalar_attributes_get_coerced_to_objects(self, User):
        user = User(website=u'www.example.com')

        assert isinstance(user.website, url.furl)



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import json_sql


@pytest.mark.usefixtures('postgresql_dsn')
class TestJSONSQL(object):

    @pytest.mark.parametrize(
        ('value', 'result'),
        (
            (1, 1),
            (14.14, 14.14),
            ({'a': 2, 'b': 'c'}, {'a': 2, 'b': 'c'}),
            (
                {'a': {'b': 'c'}},
                {'a': {'b': 'c'}}
            ),
            ({}, {}),
            ([1, 2], [1, 2]),
            ([], []),
            (
                [sa.select([sa.text('1')]).label('alias')],
                [1]
            )
        )
    )
    def test_compiled_scalars(self, connection, value, result):
        assert result == (
            connection.execute(sa.select([json_sql(value)])).fetchone()[0]
        )



import pytest
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import TSVECTOR

from sqlalchemy_utils import TSVectorType


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        search_index = sa.Column(
            TSVectorType(name, regconfig='pg_catalog.finnish')
        )

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestTSVector(object):

    def test_generates_table(self, User):
        assert 'search_index' in User.__table__.c

    @pytest.mark.usefixtures('session')
    def test_type_reflection(self, engine):
        reflected_metadata = sa.schema.MetaData()
        table = sa.schema.Table(
            'user',
            reflected_metadata,
            autoload=True,
            autoload_with=engine
        )
        assert isinstance(table.c['search_index'].type, TSVECTOR)

    def test_catalog_and_columns_as_args(self):
        type_ = TSVectorType('name', 'age', regconfig='pg_catalog.simple')
        assert type_.columns == ('name', 'age')
        assert type_.options['regconfig'] == 'pg_catalog.simple'

    def test_match(self, connection, User):
        expr = User.search_index.match(u'something')
        assert str(expr.compile(connection)) == (
            '''"user".search_index @@ to_tsquery('pg_catalog.finnish', '''
            '''%(search_index_1)s)'''
        )

    def test_concat(self, User):
        assert str(User.search_index | User.search_index) == (
            '"user".search_index || "user".search_index'
        )

    def test_match_concatenation(self, session, User):
        concat = User.search_index | User.search_index
        bind = session.bind
        assert str(concat.match('something').compile(bind)) == (
            '("user".search_index || "user".search_index) @@ '
            "to_tsquery('pg_catalog.finnish', %(param_1)s)"
        )

    def test_match_with_catalog(self, connection, User):
        expr = User.search_index.match(
            u'something',
            postgresql_regconfig='pg_catalog.simple'
        )
        assert str(expr.compile(connection)) == (
            '''"user".search_index @@ to_tsquery('pg_catalog.simple', '''
            '''%(search_index_1)s)'''
        )



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.observer import observes


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForColumn(object):

    @pytest.fixture
    def Product(self, Base):
        class Product(Base):
            __tablename__ = 'product'
            id = sa.Column(sa.Integer, primary_key=True)
            price = sa.Column(sa.Integer)

            @observes('price')
            def product_price_observer(self, price):
                self.price = price * 2
        return Product

    @pytest.fixture
    def init_models(self, Product):
        pass

    def test_simple_insert(self, session, Product):
        product = Product(price=100)
        session.add(product)
        session.flush()
        assert product.price == 200


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForColumnWithoutActualChanges(object):

    @pytest.fixture
    def Product(self, Base):
        class Product(Base):
            __tablename__ = 'product'
            id = sa.Column(sa.Integer, primary_key=True)
            price = sa.Column(sa.Integer)

            @observes('price')
            def product_price_observer(self, price):
                raise Exception('Trying to change price')
        return Product

    @pytest.fixture
    def init_models(self, Product):
        pass

    def test_only_notifies_observer_on_actual_changes(self, session, Product):
        product = Product()
        session.add(product)
        session.flush()

        with pytest.raises(Exception) as e:
            product.price = 500
            session.commit()
        assert str(e.value) == 'Trying to change price'


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForMultipleColumns(object):

    @pytest.fixture
    def Order(self, Base):
        class Order(Base):
            __tablename__ = 'order'
            id = sa.Column(sa.Integer, primary_key=True)
            unit_price = sa.Column(sa.Integer)
            amount = sa.Column(sa.Integer)
            total_price = sa.Column(sa.Integer)

            @observes('amount', 'unit_price')
            def total_price_observer(self, amount, unit_price):
                self.total_price = amount * unit_price
        return Order

    @pytest.fixture
    def init_models(self, Order):
        pass

    def test_only_notifies_observer_on_actual_changes(self, session, Order):
        order = Order()
        order.amount = 2
        order.unit_price = 10
        session.add(order)
        session.flush()

        order.amount = 1
        session.flush()
        assert order.total_price == 10

        order.unit_price = 100
        session.flush()
        assert order.total_price == 100


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForMultipleColumnsFiresOnlyOnce(object):

    @pytest.fixture
    def Order(self, Base):
        class Order(Base):
            __tablename__ = 'order'
            id = sa.Column(sa.Integer, primary_key=True)
            unit_price = sa.Column(sa.Integer)
            amount = sa.Column(sa.Integer)

            @observes('amount', 'unit_price')
            def total_price_observer(self, amount, unit_price):
                self.call_count = self.call_count + 1
        return Order

    @pytest.fixture
    def init_models(self, Order):
        pass

    def test_only_notifies_observer_on_actual_changes(self, session, Order):
        order = Order()
        order.amount = 2
        order.unit_price = 10
        order.call_count = 0
        session.add(order)
        session.flush()
        assert order.call_count == 1

        order.amount = 1
        order.unit_price = 100
        session.flush()
        assert order.call_count == 2



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import is_loaded


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article_translation'
        id = sa.Column(sa.Integer, primary_key=True)
        title = sa.orm.deferred(sa.Column(sa.String(100)))
    return Article


class TestIsLoaded(object):

    def test_loaded_property(self, Article):
        article = Article(id=1)
        assert is_loaded(article, 'id')

    def test_unloaded_property(self, Article):
        article = Article(id=4)
        assert not is_loaded(article, 'title')



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import BIT


class BitType(sa.types.TypeDecorator):
    """
    BitType offers way of saving BITs into database.
    """
    impl = sa.types.BINARY

    def __init__(self, length=1, **kwargs):
        self.length = length
        sa.types.TypeDecorator.__init__(self, **kwargs)

    def load_dialect_impl(self, dialect):
        # Use the native BIT type for drivers that has it.
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(BIT(self.length))
        elif dialect.name == 'sqlite':
            return dialect.type_descriptor(sa.String(self.length))
        else:
            return dialect.type_descriptor(type(self.impl)(self.length))



from datetime import datetime

import pytest
import sqlalchemy as sa

from sqlalchemy_utils.listeners import force_instant_defaults

force_instant_defaults()


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255), default=u'Some article')
        created_at = sa.Column(sa.DateTime, default=datetime.now)
    return Article


class TestInstantDefaultListener(object):

    def test_assigns_defaults_on_object_construction(self, Article):
        article = Article()
        assert article.name == u'Some article'

    def test_callables_as_defaults(self, Article):
        article = Article()
        assert isinstance(article.created_at, datetime)



import six
from sqlalchemy import types

from ..exceptions import ImproperlyConfigured
from .scalar_coercible import ScalarCoercible


class TimezoneType(types.TypeDecorator, ScalarCoercible):
    """
    TimezoneType provides a way for saving timezones (from either the pytz or
    the dateutil package) objects into database. TimezoneType saves timezone
    objects as strings on the way in and converts them back to objects when
    querying the database.


    ::

        from sqlalchemy_utils import TimezoneType

        class User(Base):
            __tablename__ = 'user'

            # Pass backend='pytz' to change it to use pytz (dateutil by
            # default)
            timezone = sa.Column(TimezoneType(backend='pytz'))
    """

    impl = types.Unicode(50)

    python_type = None

    def __init__(self, backend='dateutil'):
        """
        :param backend: Whether to use 'dateutil' or 'pytz' for timezones.
        """

        self.backend = backend
        if backend == 'dateutil':
            try:
                from dateutil.tz import tzfile
                from dateutil.zoneinfo import get_zonefile_instance

                self.python_type = tzfile
                self._to = get_zonefile_instance().zones.get
                self._from = lambda x: six.text_type(x._filename)

            except ImportError:
                raise ImproperlyConfigured(
                    "'python-dateutil' is required to use the "
                    "'dateutil' backend for 'TimezoneType'"
                )

        elif backend == 'pytz':
            try:
                from pytz import timezone
                from pytz.tzinfo import BaseTzInfo

                self.python_type = BaseTzInfo
                self._to = timezone
                self._from = six.text_type

            except ImportError:
                raise ImproperlyConfigured(
                    "'pytz' is required to use the 'pytz' backend "
                    "for 'TimezoneType'"
                )

        else:
            raise ImproperlyConfigured(
                "'pytz' or 'dateutil' are the backends supported for "
                "'TimezoneType'"
            )

    def _coerce(self, value):
        if value is not None and not isinstance(value, self.python_type):
            obj = self._to(value)
            if obj is None:
                raise ValueError("unknown time zone '%s'" % value)
            return obj
        return value

    def process_bind_param(self, value, dialect):
        return self._from(self._coerce(value)) if value else None

    def process_result_value(self, value, dialect):
        return self._to(value) if value else None



class TestInstrumentedList(object):
    def test_any_returns_true_if_member_has_attr_defined(
        self,
        Category,
        Article
    ):
        category = Category()
        category.articles.append(Article())
        category.articles.append(Article(name=u'some name'))
        assert category.articles.any('name')

    def test_any_returns_false_if_no_member_has_attr_defined(
        self,
        Category,
        Article
    ):
        category = Category()
        category.articles.append(Article())
        assert not category.articles.any('name')



import pytest
import six
import sqlalchemy as sa

from sqlalchemy_utils import ScalarListType


class TestScalarIntegerList(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            some_list = sa.Column(ScalarListType(int))

            def __repr__(self):
                return 'User(%r)' % self.id

        return User

    @pytest.fixture
    def init_models(self, User):
        pass

    def test_save_integer_list(self, session, User):
        user = User(
            some_list=[1, 2, 3, 4]
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.some_list == [1, 2, 3, 4]


class TestScalarUnicodeList(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            some_list = sa.Column(ScalarListType(six.text_type))

            def __repr__(self):
                return 'User(%r)' % self.id

        return User

    @pytest.fixture
    def init_models(self, User):
        pass

    def test_throws_exception_if_using_separator_in_list_values(
        self,
        session,
        User
    ):
        user = User(
            some_list=[u',']
        )

        session.add(user)
        with pytest.raises(sa.exc.StatementError) as db_err:
            session.commit()
        assert (
            "List values can't contain string ',' (its being used as "
            "separator. If you wish for scalar list values to contain "
            "these strings, use a different separator string.)"
        ) in str(db_err.value)

    def test_save_unicode_list(self, session, User):
        user = User(
            some_list=[u'1', u'2', u'3', u'4']
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.some_list == [u'1', u'2', u'3', u'4']

    def test_save_and_retrieve_empty_list(self, session, User):
        user = User(
            some_list=[]
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.some_list == []



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import table_name


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
    return Building


@pytest.fixture
def init_models(Base):
    pass


class TestTableName(object):

    def test_class(self, Building):
        assert table_name(Building) == 'building'
        del Building.__tablename__
        assert table_name(Building) == 'building'

    def test_attribute(self, Building):
        assert table_name(Building.id) == 'building'
        assert table_name(Building.name) == 'building'

    def test_target(self, Building):
        assert table_name(Building()) == 'building'



# -*- coding: utf-8 -*-
import pytest
import six

from sqlalchemy_utils import Currency, i18n


@pytest.fixture
def set_get_locale():
    i18n.get_locale = lambda: i18n.babel.Locale('en')


@pytest.mark.skipif('i18n.babel is None')
@pytest.mark.usefixtures('set_get_locale')
class TestCurrency(object):

    def test_init(self):
        assert Currency('USD') == Currency(Currency('USD'))

    def test_hashability(self):
        assert len(set([Currency('USD'), Currency('USD')])) == 1

    def test_invalid_currency_code(self):
        with pytest.raises(ValueError):
            Currency('Unknown code')

    def test_invalid_currency_code_type(self):
        with pytest.raises(TypeError):
            Currency(None)

    @pytest.mark.parametrize(
        ('code', 'name'),
        (
            ('USD', 'US Dollar'),
            ('EUR', 'Euro')
        )
    )
    def test_name_property(self, code, name):
        assert Currency(code).name == name

    @pytest.mark.parametrize(
        ('code', 'symbol'),
        (
            ('USD', u'$'),
            ('EUR', u'€')
        )
    )
    def test_symbol_property(self, code, symbol):
        assert Currency(code).symbol == symbol

    def test_equality_operator(self):
        assert Currency('USD') == 'USD'
        assert 'USD' == Currency('USD')
        assert Currency('USD') == Currency('USD')

    def test_non_equality_operator(self):
        assert Currency('USD') != 'EUR'
        assert not (Currency('USD') != 'USD')

    def test_unicode(self):
        currency = Currency('USD')
        assert six.text_type(currency) == u'USD'

    def test_str(self):
        currency = Currency('USD')
        assert str(currency) == 'USD'

    def test_representation(self):
        currency = Currency('USD')
        assert repr(currency) == "Currency('USD')"



import pytest
import sqlalchemy as sa
from flexmock import flexmock

from sqlalchemy_utils import proxy_dict, ProxyDict


@pytest.fixture
def ArticleTranslation(Base):
    class ArticleTranslation(Base):
        __tablename__ = 'article_translation'

        id = sa.Column(
            sa.Integer,
            sa.ForeignKey('article.id'),
            primary_key=True
        )
        locale = sa.Column(sa.String(10), primary_key=True)
        name = sa.Column(sa.UnicodeText)
    return ArticleTranslation


@pytest.fixture
def Article(Base, ArticleTranslation):

    class Article(Base):
        __tablename__ = 'article'

        id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
        description = sa.Column(sa.UnicodeText)
        _translations = sa.orm.relationship(
            ArticleTranslation,
            lazy='dynamic',
            cascade='all, delete-orphan',
            passive_deletes=True,
            backref=sa.orm.backref('parent'),
        )

        @property
        def translations(self):
            return proxy_dict(
                self,
                '_translations',
                ArticleTranslation.locale
            )
    return Article


@pytest.fixture
def init_models(ArticleTranslation, Article):
    pass


class TestProxyDict(object):

    def test_access_key_for_pending_parent(self, session, Article):
        article = Article()
        session.add(article)
        assert article.translations['en']

    def test_access_key_for_transient_parent(self, Article):
        article = Article()
        assert article.translations['en']

    def test_cache(self, session, Article):
        article = Article()
        (
            flexmock(ProxyDict)
            .should_receive('fetch')
            .once()
        )
        session.add(article)
        session.commit()
        article.translations['en']
        article.translations['en']

    def test_set_updates_cache(self, session, Article, ArticleTranslation):
        article = Article()
        (
            flexmock(ProxyDict)
            .should_receive('fetch')
            .once()
        )
        session.add(article)
        session.commit()
        article.translations['en']
        article.translations['en'] = ArticleTranslation(
            locale='en',
            name=u'something'
        )
        article.translations['en']

    def test_contains_efficiency(self, connection, session, Article):
        article = Article()
        session.add(article)
        session.commit()
        article.id
        query_count = connection.query_count
        'en' in article.translations
        'en' in article.translations
        'en' in article.translations
        assert connection.query_count == query_count + 1

    def test_getitem_with_none_value_in_cache(self, session, Article):
        article = Article()
        session.add(article)
        session.commit()
        article.id
        'en' in article.translations
        assert article.translations['en']

    def test_contains(self, Article):
        article = Article()
        assert 'en' not in article.translations
        # does not auto-append new translation
        assert 'en' not in article.translations

    def test_committing_session_empties_proxy_dict_cache(
        self,
        session,
        Article
    ):
        article = Article()
        (
            flexmock(ProxyDict)
            .should_receive('fetch')
            .twice()
        )
        session.add(article)
        session.commit()
        article.translations['en']
        session.commit()
        article.translations['en']



import pytest
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import ARRAY

from sqlalchemy_utils import (
    assert_max_length,
    assert_max_value,
    assert_min_value,
    assert_non_nullable,
    assert_nullable
)


@pytest.fixture()
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column('_id', sa.Integer, primary_key=True)
        name = sa.Column('_name', sa.String(20))
        age = sa.Column('_age', sa.Integer, nullable=False)
        email = sa.Column(
            '_email', sa.String(200), nullable=False, unique=True
        )
        fav_numbers = sa.Column('_fav_numbers', ARRAY(sa.Integer))

        __table_args__ = (
            sa.CheckConstraint(sa.and_(age >= 0, age <= 150)),
            sa.CheckConstraint(
                sa.and_(
                    sa.func.array_length(fav_numbers, 1) <= 8
                )
            )
        )
    return User


@pytest.fixture()
def user(User, session):
    user = User(
        name='Someone',
        email='someone@example.com',
        age=15,
        fav_numbers=[1, 2, 3]
    )
    session.add(user)
    session.commit()
    return user


@pytest.mark.usefixtures('postgresql_dsn')
class TestAssertMaxLengthWithArray(object):

    def test_with_max_length(self, user):
        assert_max_length(user, 'fav_numbers', 8)
        assert_max_length(user, 'fav_numbers', 8)

    def test_smaller_than_max_length(self, user):
        with pytest.raises(AssertionError):
            assert_max_length(user, 'fav_numbers', 7)
        with pytest.raises(AssertionError):
            assert_max_length(user, 'fav_numbers', 7)

    def test_bigger_than_max_length(self, user):
        with pytest.raises(AssertionError):
            assert_max_length(user, 'fav_numbers', 9)
        with pytest.raises(AssertionError):
            assert_max_length(user, 'fav_numbers', 9)


@pytest.mark.usefixtures('postgresql_dsn')
class TestAssertNonNullable(object):

    def test_non_nullable_column(self, user):
        # Test everything twice so that session gets rolled back properly
        assert_non_nullable(user, 'age')
        assert_non_nullable(user, 'age')

    def test_nullable_column(self, user):
        with pytest.raises(AssertionError):
            assert_non_nullable(user, 'name')
        with pytest.raises(AssertionError):
            assert_non_nullable(user, 'name')


@pytest.mark.usefixtures('postgresql_dsn')
class TestAssertNullable(object):

    def test_nullable_column(self, user):
        assert_nullable(user, 'name')
        assert_nullable(user, 'name')

    def test_non_nullable_column(self, user):
        with pytest.raises(AssertionError):
            assert_nullable(user, 'age')
        with pytest.raises(AssertionError):
            assert_nullable(user, 'age')


@pytest.mark.usefixtures('postgresql_dsn')
class TestAssertMaxLength(object):

    def test_with_max_length(self, user):
        assert_max_length(user, 'name', 20)
        assert_max_length(user, 'name', 20)

    def test_with_non_nullable_column(self, user):
        assert_max_length(user, 'email', 200)
        assert_max_length(user, 'email', 200)

    def test_smaller_than_max_length(self, user):
        with pytest.raises(AssertionError):
            assert_max_length(user, 'name', 19)
        with pytest.raises(AssertionError):
            assert_max_length(user, 'name', 19)

    def test_bigger_than_max_length(self, user):
        with pytest.raises(AssertionError):
            assert_max_length(user, 'name', 21)
        with pytest.raises(AssertionError):
            assert_max_length(user, 'name', 21)


@pytest.mark.usefixtures('postgresql_dsn')
class TestAssertMinValue(object):

    def test_with_min_value(self, user):
        assert_min_value(user, 'age', 0)
        assert_min_value(user, 'age', 0)

    def test_smaller_than_min_value(self, user):
        with pytest.raises(AssertionError):
            assert_min_value(user, 'age', -1)
        with pytest.raises(AssertionError):
            assert_min_value(user, 'age', -1)

    def test_bigger_than_min_value(self, user):
        with pytest.raises(AssertionError):
            assert_min_value(user, 'age', 1)
        with pytest.raises(AssertionError):
            assert_min_value(user, 'age', 1)


@pytest.mark.usefixtures('postgresql_dsn')
class TestAssertMaxValue(object):

    def test_with_min_value(self, user):
        assert_max_value(user, 'age', 150)
        assert_max_value(user, 'age', 150)

    def test_smaller_than_max_value(self, user):
        with pytest.raises(AssertionError):
            assert_max_value(user, 'age', 149)
        with pytest.raises(AssertionError):
            assert_max_value(user, 'age', 149)

    def test_bigger_than_max_value(self, user):
        with pytest.raises(AssertionError):
            assert_max_value(user, 'age', 151)
        with pytest.raises(AssertionError):
            assert_max_value(user, 'age', 151)



import pytest
import sqlalchemy as sa
from flexmock import flexmock

from sqlalchemy_utils import create_database, database_exists, drop_database

pymysql = None
try:
    import pymysql  # noqa
except ImportError:
    pass


class DatabaseTest(object):
    def test_create_and_drop(self, dsn):
        assert not database_exists(dsn)
        create_database(dsn)
        assert database_exists(dsn)
        drop_database(dsn)
        assert not database_exists(dsn)


@pytest.mark.usefixtures('sqlite_memory_dsn')
class TestDatabaseSQLiteMemory(object):

    def test_exists_memory(self, dsn):
        assert database_exists(dsn)


@pytest.mark.usefixtures('sqlite_none_database_dsn')
class TestDatabaseSQLiteMemoryNoDatabaseString(object):
    def test_exists_memory_none_database(self, sqlite_none_database_dsn):
        assert database_exists(sqlite_none_database_dsn)


@pytest.mark.usefixtures('sqlite_file_dsn')
class TestDatabaseSQLiteFile(DatabaseTest):
    def test_existing_non_sqlite_file(self, dsn):
        database = sa.engine.url.make_url(dsn).database
        open(database, 'w').close()
        self.test_create_and_drop(dsn)


@pytest.mark.skipif('pymysql is None')
@pytest.mark.usefixtures('mysql_dsn')
class TestDatabaseMySQL(DatabaseTest):

    @pytest.fixture
    def db_name(self):
        return 'db_test_sqlalchemy_util'


@pytest.mark.skipif('pymysql is None')
@pytest.mark.usefixtures('mysql_dsn')
class TestDatabaseMySQLWithQuotedName(DatabaseTest):

    @pytest.fixture
    def db_name(self):
        return 'db_test_sqlalchemy-util'


@pytest.mark.usefixtures('postgresql_dsn')
class TestDatabasePostgres(DatabaseTest):

    @pytest.fixture
    def db_name(self):
        return 'db_test_sqlalchemy_util'

    def test_template(self, postgresql_db_user):
        (
            flexmock(sa.engine.Engine)
            .should_receive('execute')
            .with_args(
                "CREATE DATABASE db_test_sqlalchemy_util ENCODING 'utf8' "
                "TEMPLATE my_template"
            )
        )
        dsn = 'postgresql://{0}@localhost/db_test_sqlalchemy_util'.format(
            postgresql_db_user
        )
        create_database(dsn, template='my_template')


class TestDatabasePostgresPg8000(DatabaseTest):

    @pytest.fixture
    def dsn(self, postgresql_db_user):
        return 'postgresql+pg8000://{0}@localhost/{1}'.format(
            postgresql_db_user,
            'db_to_test_create_and_drop_via_pg8000_driver'
        )


@pytest.mark.usefixtures('postgresql_dsn')
class TestDatabasePostgresWithQuotedName(DatabaseTest):

    @pytest.fixture
    def db_name(self):
        return 'db_test_sqlalchemy-util'

    def test_template(self, postgresql_db_user):
        (
            flexmock(sa.engine.Engine)
            .should_receive('execute')
            .with_args(
                '''CREATE DATABASE "db_test_sqlalchemy-util"'''
                " ENCODING 'utf8' "
                'TEMPLATE "my-template"'
            )
        )
        dsn = 'postgresql://{0}@localhost/db_test_sqlalchemy-util'.format(
            postgresql_db_user
        )
        create_database(dsn, template='my-template')


class TestDatabasePostgresCreateDatabaseCloseConnection(object):
    def test_create_database_twice(self, postgresql_db_user):
        dsn_list = [
            'postgresql://{0}@localhost/db_test_sqlalchemy-util-a'.format(
                postgresql_db_user
            ),
            'postgres://{0}@localhost/db_test_sqlalchemy-util-b'.format(
                postgresql_db_user
            ),
        ]
        for dsn_item in dsn_list:
            assert not database_exists(dsn_item)
            create_database(dsn_item, template="template1")
            assert database_exists(dsn_item)
        for dsn_item in dsn_list:
            drop_database(dsn_item)
            assert not database_exists(dsn_item)


@pytest.mark.usefixtures('mssql_dsn')
class TestDatabaseMssql(DatabaseTest):

    @pytest.fixture
    def db_name(self):
        pytest.importorskip('pyodbc')
        return 'db_test_sqlalchemy_util'



import pytest
import six
import sqlalchemy as sa

from sqlalchemy_utils import generic_relationship


@pytest.fixture
def Employee(Base):
    class Employee(Base):
        __tablename__ = 'employee'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String(50))
        type = sa.Column(sa.String(20))

        __mapper_args__ = {
            'polymorphic_on': type,
            'polymorphic_identity': 'employee'
        }
    return Employee


@pytest.fixture
def Manager(Employee):
    class Manager(Employee):
        __mapper_args__ = {
            'polymorphic_identity': 'manager'
        }
    return Manager


@pytest.fixture
def Engineer(Employee):
    class Engineer(Employee):
        __mapper_args__ = {
            'polymorphic_identity': 'engineer'
        }
    return Engineer


@pytest.fixture
def Event(Base):
    class Event(Base):
        __tablename__ = 'event'
        id = sa.Column(sa.Integer, primary_key=True)

        object_type = sa.Column(sa.Unicode(255))
        object_id = sa.Column(sa.Integer, nullable=False)

        object = generic_relationship(object_type, object_id)
    return Event


@pytest.fixture
def init_models(Employee, Manager, Engineer, Event):
    pass


class TestGenericRelationship(object):

    def test_set_as_none(self, Event):
        event = Event()
        event.object = None
        assert event.object is None

    def test_set_manual_and_get(self, session, Manager, Event):
        manager = Manager()

        session.add(manager)
        session.commit()

        event = Event()
        event.object_id = manager.id
        event.object_type = six.text_type(type(manager).__name__)

        assert event.object is None

        session.add(event)
        session.commit()

        assert event.object == manager

    def test_set_and_get(self, session, Manager, Event):
        manager = Manager()

        session.add(manager)
        session.commit()

        event = Event(object=manager)

        assert event.object_id == manager.id
        assert event.object_type == type(manager).__name__

        session.add(event)
        session.commit()

        assert event.object == manager

    def test_compare_instance(self, session, Manager, Event):
        manager1 = Manager()
        manager2 = Manager()

        session.add_all([manager1, manager2])
        session.commit()

        event = Event(object=manager1)

        session.add(event)
        session.commit()

        assert event.object == manager1
        assert event.object != manager2

    def test_compare_query(self, session, Manager, Event):
        manager1 = Manager()
        manager2 = Manager()

        session.add_all([manager1, manager2])
        session.commit()

        event = Event(object=manager1)

        session.add(event)
        session.commit()

        q = session.query(Event)
        assert q.filter_by(object=manager1).first() is not None
        assert q.filter_by(object=manager2).first() is None
        assert q.filter(Event.object == manager2).first() is None

    def test_compare_not_query(self, session, Manager, Event):
        manager1 = Manager()
        manager2 = Manager()

        session.add_all([manager1, manager2])
        session.commit()

        event = Event(object=manager1)

        session.add(event)
        session.commit()

        q = session.query(Event)
        assert q.filter(Event.object != manager2).first() is not None

    def test_compare_type(self, session, Manager, Event):
        manager1 = Manager()
        manager2 = Manager()

        session.add_all([manager1, manager2])
        session.commit()

        event1 = Event(object=manager1)
        event2 = Event(object=manager2)

        session.add_all([event1, event2])
        session.commit()

        statement = Event.object.is_type(Manager)
        q = session.query(Event).filter(statement)
        assert q.first() is not None

    def test_compare_super_type(self, session, Manager, Event, Employee):
        manager1 = Manager()
        manager2 = Manager()

        session.add_all([manager1, manager2])
        session.commit()

        event1 = Event(object=manager1)
        event2 = Event(object=manager2)

        session.add_all([event1, event2])
        session.commit()

        statement = Event.object.is_type(Employee)
        q = session.query(Event).filter(statement)
        assert q.first() is not None



from itertools import chain

import pytest
import sqlalchemy as sa

from sqlalchemy_utils.functions import non_indexed_foreign_keys


class TestFindNonIndexedForeignKeys(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return User

    @pytest.fixture
    def Category(self, Base):
        class Category(Base):
            __tablename__ = 'category'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return Category

    @pytest.fixture
    def Article(self, Base, User, Category):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
            author_id = sa.Column(
                sa.Integer, sa.ForeignKey(User.id), index=True
            )
            category_id = sa.Column(sa.Integer, sa.ForeignKey(Category.id))

            category = sa.orm.relationship(
                Category,
                primaryjoin=category_id == Category.id,
                backref=sa.orm.backref(
                    'articles',
                )
            )
        return Article

    @pytest.fixture
    def init_models(self, User, Category, Article):
        pass

    def test_finds_all_non_indexed_fks(self, session, Base, engine):
        fks = non_indexed_foreign_keys(Base.metadata, engine)
        assert (
            'article' in
            fks
        )
        column_names = list(chain(
            *(
                names for names in (
                    fk.columns.keys()
                    for fk in fks['article']
                )
            )
        ))
        assert 'category_id' in column_names
        assert 'author_id' not in column_names



import pytest

from sqlalchemy_utils.relationships import chained_join

from ..mixins import (
    ThreeLevelDeepManyToMany,
    ThreeLevelDeepOneToMany,
    ThreeLevelDeepOneToOne
)


@pytest.mark.usefixtures('postgresql_dsn')
class TestChainedJoinFoDeepToManyToMany(ThreeLevelDeepManyToMany):

    def test_simple_join(self, Catalog):
        assert str(chained_join(Catalog.categories)) == (
            'catalog_category JOIN category ON '
            'category._id = catalog_category.category_id'
        )

    def test_two_relations(self, Catalog, Category):
        sql = chained_join(
            Catalog.categories,
            Category.sub_categories
        )
        assert str(sql) == (
            'catalog_category JOIN category ON category._id = '
            'catalog_category.category_id JOIN category_subcategory ON '
            'category._id = category_subcategory.category_id JOIN '
            'sub_category ON sub_category._id = '
            'category_subcategory.subcategory_id'
        )

    def test_three_relations(self, Catalog, Category, SubCategory):
        sql = chained_join(
            Catalog.categories,
            Category.sub_categories,
            SubCategory.products
        )
        assert str(sql) == (
            'catalog_category JOIN category ON category._id = '
            'catalog_category.category_id JOIN category_subcategory ON '
            'category._id = category_subcategory.category_id JOIN sub_category'
            ' ON sub_category._id = category_subcategory.subcategory_id JOIN '
            'subcategory_product ON sub_category._id = '
            'subcategory_product.subcategory_id JOIN product ON product._id ='
            ' subcategory_product.product_id'
        )


@pytest.mark.usefixtures('postgresql_dsn')
class TestChainedJoinForDeepOneToMany(ThreeLevelDeepOneToMany):

    def test_simple_join(self, Catalog):
        assert str(chained_join(Catalog.categories)) == 'category'

    def test_two_relations(self, Catalog, Category):
        sql = chained_join(
            Catalog.categories,
            Category.sub_categories
        )
        assert str(sql) == (
            'category JOIN sub_category ON category._id = '
            'sub_category._category_id'
        )

    def test_three_relations(self, Catalog, Category, SubCategory):
        sql = chained_join(
            Catalog.categories,
            Category.sub_categories,
            SubCategory.products
        )
        assert str(sql) == (
            'category JOIN sub_category ON category._id = '
            'sub_category._category_id JOIN product ON sub_category._id = '
            'product._sub_category_id'
        )


@pytest.mark.usefixtures('postgresql_dsn')
class TestChainedJoinForDeepOneToOne(ThreeLevelDeepOneToOne):

    def test_simple_join(self, Catalog):
        assert str(chained_join(Catalog.category)) == 'category'

    def test_two_relations(self, Catalog, Category):
        sql = chained_join(
            Catalog.category,
            Category.sub_category
        )
        assert str(sql) == (
            'category JOIN sub_category ON category._id = '
            'sub_category._category_id'
        )

    def test_three_relations(self, Catalog, Category, SubCategory):
        sql = chained_join(
            Catalog.category,
            Category.sub_category,
            SubCategory.product
        )
        assert str(sql) == (
            'category JOIN sub_category ON category._id = '
            'sub_category._category_id JOIN product ON sub_category._id = '
            'product._sub_category_id'
        )



from decimal import Decimal

import pytest
import sqlalchemy as sa

from sqlalchemy_utils.aggregates import aggregated


@pytest.fixture
def Catalog(Base):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @aggregated(
            'categories.products',
            sa.Column(sa.Integer, default=0)
        )
        def product_count(self):
            return sa.func.count('1')

        categories = sa.orm.relationship('Category', backref='catalog')
    return Catalog


@pytest.fixture
def Category(Base):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))

        products = sa.orm.relationship('Product', backref='category')
    return Category


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        price = sa.Column(sa.Numeric)

        category_id = sa.Column(sa.Integer, sa.ForeignKey('category.id'))
    return Product


@pytest.fixture
def init_models(Catalog, Category, Product):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestAggregateOneToManyAndOneToMany(object):

    def test_assigns_aggregates(self, session, Category, Catalog, Product):
        category = Category(name=u'Some category')
        catalog = Catalog(
            categories=[category]
        )
        catalog.name = u'Some catalog'
        session.add(catalog)
        session.commit()
        product = Product(
            name=u'Some product',
            price=Decimal('1000'),
            category=category
        )
        session.add(product)
        session.commit()
        session.refresh(catalog)
        assert catalog.product_count == 1



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import EmailType


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        email = sa.Column(EmailType)
        short_email = sa.Column(EmailType(length=70))

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


class TestEmailType(object):
    def test_saves_email_as_lowercased(self, session, User):
        user = User(email=u'Someone@example.com')

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.email == u'someone@example.com'

    def test_literal_param(self, session, User):
        clause = User.email == 'Someone@example.com'
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == '"user".email = lower(\'Someone@example.com\')'

    def test_custom_length(self, session, User):
        assert User.short_email.type.impl.length == 70



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import aggregated


@pytest.fixture
def Catalog(Base):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)

        @aggregated(
            'categories.sub_categories.products',
            sa.Column(sa.Integer, default=0)
        )
        def product_count(self):
            return sa.func.count('1')

        categories = sa.orm.relationship('Category', backref='catalog')
    return Catalog


@pytest.fixture
def Category(Base):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))

        sub_categories = sa.orm.relationship(
            'SubCategory', backref='category'
        )
    return Category


@pytest.fixture
def SubCategory(Base):
    class SubCategory(Base):
        __tablename__ = 'sub_category'
        id = sa.Column(sa.Integer, primary_key=True)
        category_id = sa.Column(sa.Integer, sa.ForeignKey('category.id'))
        products = sa.orm.relationship('Product', backref='sub_category')
    return SubCategory


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        price = sa.Column(sa.Numeric)

        sub_category_id = sa.Column(
            sa.Integer, sa.ForeignKey('sub_category.id')
        )
    return Product


@pytest.fixture
def init_models(Catalog, Category, SubCategory, Product):
    pass


@pytest.fixture
def catalog_factory(Product, SubCategory, Category, Catalog, session):
    def catalog_factory():
        product = Product()
        sub_category = SubCategory(
            products=[product]
        )
        category = Category(sub_categories=[sub_category])
        catalog = Catalog(categories=[category])
        session.add(catalog)
        return catalog
    return catalog_factory


@pytest.mark.usefixtures('postgresql_dsn')
class Test3LevelDeepOneToMany(object):

    def test_assigns_aggregates(self, session, catalog_factory):
        catalog = catalog_factory()
        session.commit()
        session.refresh(catalog)
        assert catalog.product_count == 1

    def catalog_factory(
        self,
        session,
        Product,
        SubCategory,
        Category,
        Catalog
    ):
        product = Product()
        sub_category = SubCategory(
            products=[product]
        )
        category = Category(sub_categories=[sub_category])
        catalog = Catalog(categories=[category])
        session.add(catalog)
        return catalog

    def test_only_updates_affected_aggregates(
        self,
        session,
        catalog_factory,
        Product
    ):
        catalog = catalog_factory()
        catalog2 = catalog_factory()
        session.commit()

        # force set catalog2 product_count to zero in order to check if it gets
        # updated when the other catalog's product count gets updated
        session.execute(
            'UPDATE catalog SET product_count = 0 WHERE id = %d'
            % catalog2.id
        )

        catalog.categories[0].sub_categories[0].products.append(
            Product()
        )
        session.commit()
        session.refresh(catalog)
        session.refresh(catalog2)
        assert catalog.product_count == 2
        assert catalog2.product_count == 0



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_mapper


class TestGetMapper(object):

    @pytest.fixture
    def Building(self, Base):
        class Building(Base):
            __tablename__ = 'building'
            id = sa.Column(sa.Integer, primary_key=True)
        return Building

    def test_table(self, Building):
        assert get_mapper(Building.__table__) == sa.inspect(Building)

    def test_declarative_class(self, Building):
        assert (
            get_mapper(Building) ==
            sa.inspect(Building)
        )

    def test_declarative_object(self, Building):
        assert (
            get_mapper(Building()) ==
            sa.inspect(Building)
        )

    def test_mapper(self, Building):
        assert (
            get_mapper(Building.__mapper__) ==
            sa.inspect(Building)
        )

    def test_class_alias(self, Building):
        assert (
            get_mapper(sa.orm.aliased(Building)) ==
            sa.inspect(Building)
        )

    def test_instrumented_attribute(self, Building):
        assert (
            get_mapper(Building.id) == sa.inspect(Building)
        )

    def test_table_alias(self, Building):
        alias = sa.orm.aliased(Building.__table__)
        assert (
            get_mapper(alias) ==
            sa.inspect(Building)
        )

    def test_column(self, Building):
        assert (
            get_mapper(Building.__table__.c.id) ==
            sa.inspect(Building)
        )

    def test_column_of_an_alias(self, Building):
        assert (
            get_mapper(sa.orm.aliased(Building.__table__).c.id) ==
            sa.inspect(Building)
        )


class TestGetMapperWithQueryEntities(object):

    @pytest.fixture
    def Building(self, Base):
        class Building(Base):
            __tablename__ = 'building'
            id = sa.Column(sa.Integer, primary_key=True)
        return Building

    @pytest.fixture
    def init_models(self, Building):
        pass

    def test_mapper_entity_with_mapper(self, session, Building):
        entity = session.query(Building.__mapper__)._entities[0]
        assert (
            get_mapper(entity) ==
            sa.inspect(Building)
        )

    def test_mapper_entity_with_class(self, session, Building):
        entity = session.query(Building)._entities[0]
        assert (
            get_mapper(entity) ==
            sa.inspect(Building)
        )

    def test_column_entity(self, session, Building):
        query = session.query(Building.id)
        assert get_mapper(query._entities[0]) == sa.inspect(Building)


class TestGetMapperWithMultipleMappersFound(object):

    @pytest.fixture
    def Building(self, Base):
        class Building(Base):
            __tablename__ = 'building'
            id = sa.Column(sa.Integer, primary_key=True)

        class BigBuilding(Building):
            pass

        return Building

    def test_table(self, Building):
        with pytest.raises(ValueError):
            get_mapper(Building.__table__)

    def test_table_alias(self, Building):
        alias = sa.orm.aliased(Building.__table__)
        with pytest.raises(ValueError):
            get_mapper(alias)


class TestGetMapperForTableWithoutMapper(object):

    @pytest.fixture
    def building(self):
        metadata = sa.MetaData()
        return sa.Table('building', metadata)

    def test_table(self, building):
        with pytest.raises(ValueError):
            get_mapper(building)

    def test_table_alias(self, building):
        alias = sa.orm.aliased(building)
        with pytest.raises(ValueError):
            get_mapper(alias)



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import Currency, CurrencyType, i18n


@pytest.fixture
def set_get_locale():
    i18n.get_locale = lambda: i18n.babel.Locale('en')


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        currency = sa.Column(CurrencyType)

        def __repr__(self):
            return 'User(%r)' % self.id

    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.mark.skipif('i18n.babel is None')
class TestCurrencyType(object):

    def test_parameter_processing(self, session, User, set_get_locale):
        user = User(
            currency=Currency('USD')
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.currency.name == u'US Dollar'

    def test_scalar_attributes_get_coerced_to_objects(
        self,
        User,
        set_get_locale
    ):
        user = User(currency='USD')
        assert isinstance(user.currency, Currency)

    def test_literal_param(self, session, User):
        clause = User.currency == 'USD'
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == '"user".currency = \'USD\''



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.functions.sort_query import make_order_by_deterministic

from .. import assert_contains


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        author_id = sa.Column(sa.Integer, sa.ForeignKey('user.id'))
        author = sa.orm.relationship('User')
    return Article


@pytest.fixture
def User(Base, Article):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode)
        email = sa.Column(sa.Unicode, unique=True)

        email_lower = sa.orm.column_property(
            sa.func.lower(name)
        )

    User.article_count = sa.orm.column_property(
        sa.select([sa.func.count()], from_obj=Article)
        .where(Article.author_id == User.id)
        .label('article_count')
    )
    return User


@pytest.fixture
def init_models(Article, User):
    pass


class TestMakeOrderByDeterministic(object):

    def test_column_property(self, session, User):
        query = session.query(User).order_by(User.email_lower)
        query = make_order_by_deterministic(query)
        assert_contains('lower(user.name) AS lower_1', query)
        assert_contains('lower_1, user.id ASC', query)

    def test_unique_column(self, session, User):
        query = session.query(User).order_by(User.email)
        query = make_order_by_deterministic(query)

        assert str(query).endswith('ORDER BY user.email')

    def test_non_unique_column(self, session, User):
        query = session.query(User).order_by(User.name)
        query = make_order_by_deterministic(query)
        assert_contains('ORDER BY user.name, user.id ASC', query)

    def test_descending_order_by(self, session, User):
        query = session.query(User).order_by(
            sa.desc(User.name)
        )
        query = make_order_by_deterministic(query)
        assert_contains('ORDER BY user.name DESC, user.id DESC', query)

    def test_ascending_order_by(self, session, User):
        query = session.query(User).order_by(
            sa.asc(User.name)
        )
        query = make_order_by_deterministic(query)
        assert_contains('ORDER BY user.name ASC, user.id ASC', query)

    def test_string_order_by(self, session, User):
        query = session.query(User).order_by('name')
        query = make_order_by_deterministic(query)
        assert_contains('ORDER BY user.name, user.id ASC', query)

    def test_annotated_label(self, session, User):
        query = session.query(User).order_by(User.article_count)
        query = make_order_by_deterministic(query)
        assert_contains('article_count, user.id ASC', query)

    def test_annotated_label_with_descending_order(self, session, User):
        query = session.query(User).order_by(
            sa.desc(User.article_count)
        )
        query = make_order_by_deterministic(query)
        assert_contains('ORDER BY article_count DESC, user.id DESC', query)

    def test_query_without_order_by(self, session, User):
        query = session.query(User)
        query = make_order_by_deterministic(query)
        assert 'ORDER BY user.id' in str(query)

    def test_alias(self, session, User):
        alias = sa.orm.aliased(User.__table__)
        query = session.query(alias).order_by(alias.c.name)
        query = make_order_by_deterministic(query)
        assert str(query).endswith('ORDER BY user_1.name, user.id ASC')



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import QueryChain


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
    return User


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
    return Article


@pytest.fixture
def BlogPost(Base):
    class BlogPost(Base):
        __tablename__ = 'blog_post'
        id = sa.Column(sa.Integer, primary_key=True)
    return BlogPost


@pytest.fixture
def init_models(User, Article, BlogPost):
    pass


@pytest.fixture
def users(session, User):
    users = [User(), User()]
    session.add_all(users)
    session.commit()
    return users


@pytest.fixture
def articles(session, Article):
    articles = [Article(), Article(), Article(), Article()]
    session.add_all(articles)
    session.commit()
    return articles


@pytest.fixture
def posts(session, BlogPost):
    posts = [BlogPost(), BlogPost(), BlogPost()]
    session.add_all(posts)
    session.commit()
    return posts


@pytest.fixture
def chain(session, users, articles, posts, User, Article, BlogPost):
    return QueryChain(
        [
            session.query(User).order_by('id'),
            session.query(Article).order_by('id'),
            session.query(BlogPost).order_by('id')
        ]
    )


class TestQueryChain(object):

    def test_iter(self, chain):
        assert len(list(chain)) == 9

    def test_iter_with_limit(self, chain, users, articles):
        c = chain.limit(4)
        objects = list(c)
        assert users == objects[0:2]
        assert articles[0:2] == objects[2:]

    def test_iter_with_offset(self, chain, articles, posts):
        c = chain.offset(3)
        objects = list(c)
        assert articles[1:] + posts == objects

    def test_iter_with_limit_and_offset(self, chain, articles, posts):
        c = chain.offset(3).limit(4)
        objects = list(c)
        assert articles[1:] + posts[0:1] == objects

    def test_iter_with_offset_spanning_multiple_queries(self, chain, posts):
        c = chain.offset(7)
        objects = list(c)
        assert posts[1:] == objects

    def test_repr(self, chain):
        assert repr(chain) == '<QueryChain at 0x%x>' % id(chain)

    def test_getitem_with_slice(self, chain):
        c = chain[1:]
        assert c._offset == 1
        assert c._limit is None

    def test_getitem_with_single_key(self, chain, articles):
        article = chain[2]
        assert article == articles[0]

    def test_count(self, chain):
        assert chain.count() == 9



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_class_by_table


class TestGetClassByTableWithJoinedTableInheritance(object):

    @pytest.fixture
    def Entity(self, Base):
        class Entity(Base):
            __tablename__ = 'entity'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String)
            type = sa.Column(sa.String)
            __mapper_args__ = {
                'polymorphic_on': type,
                'polymorphic_identity': 'entity'
            }
        return Entity

    @pytest.fixture
    def User(self, Entity):
        class User(Entity):
            __tablename__ = 'user'
            id = sa.Column(
                sa.Integer,
                sa.ForeignKey(Entity.id, ondelete='CASCADE'),
                primary_key=True
            )
            __mapper_args__ = {
                'polymorphic_identity': 'user'
            }
        return User

    def test_returns_class(self, Base, User, Entity):
        assert get_class_by_table(Base, User.__table__) == User
        assert get_class_by_table(
            Base,
            Entity.__table__
        ) == Entity

    def test_table_with_no_associated_class(self, Base):
        table = sa.Table(
            'some_table',
            Base.metadata,
            sa.Column('id', sa.Integer)
        )
        assert get_class_by_table(Base, table) is None


class TestGetClassByTableWithSingleTableInheritance(object):

    @pytest.fixture
    def Entity(self, Base):
        class Entity(Base):
            __tablename__ = 'entity'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String)
            type = sa.Column(sa.String)
            __mapper_args__ = {
                'polymorphic_on': type,
                'polymorphic_identity': 'entity'
            }
        return Entity

    @pytest.fixture
    def User(self, Entity):
        class User(Entity):
            __mapper_args__ = {
                'polymorphic_identity': 'user'
            }
        return User

    def test_multiple_classes_without_data_parameter(self, Base, Entity, User):
        with pytest.raises(ValueError):
            assert get_class_by_table(
                Base,
                Entity.__table__
            )

    def test_multiple_classes_with_data_parameter(self, Base, Entity, User):
        assert get_class_by_table(
            Base,
            Entity.__table__,
            {'type': 'entity'}
        ) == Entity
        assert get_class_by_table(
            Base,
            Entity.__table__,
            {'type': 'user'}
        ) == User

    def test_multiple_classes_with_bogus_data(self, Base, Entity, User):
        with pytest.raises(ValueError):
            assert get_class_by_table(
                Base,
                Entity.__table__,
                {'type': 'unknown'}
            )



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.aggregates import aggregated


@pytest.fixture
def User(Base):
    user_group = sa.Table(
        'user_group',
        Base.metadata,
        sa.Column('user_id', sa.Integer, sa.ForeignKey('user.id')),
        sa.Column('group_id', sa.Integer, sa.ForeignKey('group.id'))
    )

    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @aggregated('groups', sa.Column(sa.Integer, default=0))
        def group_count(self):
            return sa.func.count('1')

        groups = sa.orm.relationship(
            'Group',
            backref='users',
            secondary=user_group
        )
    return User


@pytest.fixture
def Group(Base):
    class Group(Base):
        __tablename__ = 'group'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
    return Group


@pytest.fixture
def init_models(User, Group):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestAggregatesWithManyToManyRelationships(object):

    def test_assigns_aggregates_on_insert(self, session, User, Group):
        user = User(
            name=u'John Matrix'
        )
        session.add(user)
        session.commit()
        group = Group(
            name=u'Some group',
            users=[user]
        )
        session.add(group)
        session.commit()
        session.refresh(user)
        assert user.group_count == 1

    def test_updates_aggregates_on_delete(self, session, User, Group):
        user = User(
            name=u'John Matrix'
        )
        session.add(user)
        session.commit()
        group = Group(
            name=u'Some group',
            users=[user]
        )
        session.add(group)
        session.commit()
        session.refresh(user)
        user.groups = []
        session.commit()
        session.refresh(user)
        assert user.group_count == 0



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_query_entities


@pytest.fixture
def TextItem(Base):
    class TextItem(Base):
        __tablename__ = 'text_item'
        id = sa.Column(sa.Integer, primary_key=True)

        type = sa.Column(sa.Unicode(255))

        __mapper_args__ = {
            'polymorphic_on': type,
        }
    return TextItem


@pytest.fixture
def Article(TextItem):
    class Article(TextItem):
        __tablename__ = 'article'
        id = sa.Column(
            sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
        )
        category = sa.Column(sa.Unicode(255))
        __mapper_args__ = {
            'polymorphic_identity': u'article'
        }
    return Article


@pytest.fixture
def BlogPost(TextItem):
    class BlogPost(TextItem):
        __tablename__ = 'blog_post'
        id = sa.Column(
            sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
        )
        __mapper_args__ = {
            'polymorphic_identity': u'blog_post'
        }
    return BlogPost


@pytest.fixture
def init_models(TextItem, Article, BlogPost):
    pass


class TestGetQueryEntities(object):

    def test_mapper(self, session, TextItem):
        query = session.query(sa.inspect(TextItem))
        assert get_query_entities(query) == [TextItem]

    def test_entity(self, session, TextItem):
        query = session.query(TextItem)
        assert get_query_entities(query) == [TextItem]

    def test_instrumented_attribute(self, session, TextItem):
        query = session.query(TextItem.id)
        assert get_query_entities(query) == [TextItem]

    def test_column(self, session, TextItem):
        query = session.query(TextItem.__table__.c.id)
        assert get_query_entities(query) == [TextItem.__table__]

    def test_aliased_selectable(self, session, TextItem, BlogPost):
        selectable = sa.orm.with_polymorphic(TextItem, [BlogPost])
        query = session.query(selectable)
        assert get_query_entities(query) == [selectable]

    def test_joined_entity(self, session, TextItem, BlogPost):
        query = session.query(TextItem).join(
            BlogPost, BlogPost.id == TextItem.id
        )
        assert get_query_entities(query) == [
            TextItem, sa.inspect(BlogPost)
        ]

    def test_joined_aliased_entity(self, session, TextItem, BlogPost):
        alias = sa.orm.aliased(BlogPost)

        query = session.query(TextItem).join(
            alias, alias.id == TextItem.id
        )
        assert get_query_entities(query) == [TextItem, alias]

    def test_column_entity_with_label(self, session, Article):
        query = session.query(Article.id.label('id'))
        assert get_query_entities(query) == [Article]

    def test_with_subquery(self, session, Article):
        number_of_articles = (
            sa.select(
                [sa.func.count(Article.id)],
            )
            .select_from(
                Article.__table__
            )
        ).label('number_of_articles')

        query = session.query(Article, number_of_articles)
        assert get_query_entities(query) == [
            Article,
            number_of_articles
        ]

    def test_aliased_entity(self, session, Article):
        alias = sa.orm.aliased(Article)
        query = session.query(alias)
        assert get_query_entities(query) == [alias]



import pytest
import six
import sqlalchemy as sa

from sqlalchemy_utils.types import ip_address


@pytest.fixture
def Visitor(Base):
    class Visitor(Base):
        __tablename__ = 'document'
        id = sa.Column(sa.Integer, primary_key=True)
        ip_address = sa.Column(ip_address.IPAddressType)

        def __repr__(self):
            return 'Visitor(%r)' % self.id
    return Visitor


@pytest.fixture
def init_models(Visitor):
    pass


@pytest.mark.skipif('ip_address.ip_address is None')
class TestIPAddressType(object):

    def test_parameter_processing(self, session, Visitor):
        visitor = Visitor(
            ip_address=u'111.111.111.111'
        )

        session.add(visitor)
        session.commit()

        visitor = session.query(Visitor).first()
        assert six.text_type(visitor.ip_address) == u'111.111.111.111'



from datetime import datetime

import pytest
import sqlalchemy as sa
from dateutil import tz

from sqlalchemy_utils.types import arrow


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        created_at = sa.Column(arrow.ArrowType)
        published_at = sa.Column(arrow.ArrowType(timezone=True))
        published_at_dt = sa.Column(sa.DateTime(timezone=True))
    return Article


@pytest.fixture
def init_models(Article):
    pass


@pytest.mark.skipif('arrow.arrow is None')
class TestArrowDateTimeType(object):

    def test_parameter_processing(self, session, Article):
        article = Article(
            created_at=arrow.arrow.get(datetime(2000, 11, 1))
        )

        session.add(article)
        session.commit()

        article = session.query(Article).first()
        assert article.created_at.datetime

    def test_string_coercion(self, Article):
        article = Article(
            created_at='2013-01-01'
        )
        assert article.created_at.year == 2013

    def test_utc(self, session, Article):
        time = arrow.arrow.utcnow()
        article = Article(created_at=time)
        session.add(article)
        assert article.created_at == time
        session.commit()
        assert article.created_at == time

    def test_other_tz(self, session, Article):
        time = arrow.arrow.utcnow()
        local = time.to('US/Pacific')
        article = Article(created_at=local)
        session.add(article)
        assert article.created_at == time == local
        session.commit()
        assert article.created_at == time

    def test_literal_param(self, session, Article):
        clause = Article.created_at > '2015-01-01'
        compiled = str(clause.compile(compile_kwargs={"literal_binds": True}))
        assert compiled == 'article.created_at > 2015-01-01'

    @pytest.mark.usefixtures('postgresql_dsn')
    def test_timezone(self, session, Article):
        timezone = tz.gettz('Europe/Stockholm')
        dt = arrow.arrow.get(datetime(2015, 1, 1, 15, 30, 45), timezone)
        article = Article(published_at=dt, published_at_dt=dt.datetime)

        session.add(article)
        session.commit()
        session.expunge_all()

        item = session.query(Article).one()
        assert item.published_at.datetime == item.published_at_dt
        assert item.published_at.to(timezone) == dt



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_primary_keys

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column('_id', sa.Integer, primary_key=True)
        name = sa.Column('_name', sa.Unicode(255))
    return Building


class TestGetPrimaryKeys(object):

    def test_table(self, Building):
        assert get_primary_keys(Building.__table__) == OrderedDict({
            '_id': Building.__table__.c._id
        })

    def test_declarative_class(self, Building):
        assert get_primary_keys(Building) == OrderedDict({
            'id': Building.__table__.c._id
        })

    def test_declarative_object(self, Building):
        assert get_primary_keys(Building()) == OrderedDict({
            'id': Building.__table__.c._id
        })

    def test_class_alias(self, Building):
        alias = sa.orm.aliased(Building)
        assert get_primary_keys(alias) == OrderedDict({
            'id': Building.__table__.c._id
        })

    def test_table_alias(self, Building):
        alias = sa.orm.aliased(Building.__table__)
        assert get_primary_keys(alias) == OrderedDict({
            '_id': alias.c._id
        })



import pytest
import six
from flexmock import flexmock

from sqlalchemy_utils import i18n
from sqlalchemy_utils.primitives import WeekDay, WeekDays


@pytest.fixture
def set_get_locale():
    i18n.get_locale = lambda: i18n.babel.Locale('fi')


@pytest.mark.skipif('i18n.babel is None')
@pytest.mark.usefixtures('set_get_locale')
class TestWeekDay(object):

    def test_constructor_with_valid_index(self):
        day = WeekDay(1)
        assert day.index == 1

    @pytest.mark.parametrize('index', [-1, 7])
    def test_constructor_with_invalid_index(self, index):
        with pytest.raises(ValueError):
            WeekDay(index)

    def test_equality_with_equal_week_day(self):
        day = WeekDay(1)
        day2 = WeekDay(1)
        assert day == day2

    def test_equality_with_unequal_week_day(self):
        day = WeekDay(1)
        day2 = WeekDay(2)
        assert day != day2

    def test_equality_with_unsupported_comparison(self):
        day = WeekDay(1)
        assert day != 'foobar'

    def test_hash_is_equal_to_index_hash(self):
        day = WeekDay(1)
        assert hash(day) == hash(day.index)

    def test_representation(self):
        day = WeekDay(1)
        assert repr(day) == "WeekDay(1)"

    @pytest.mark.parametrize(
        ('index', 'first_week_day', 'position'),
        [
            (0, 0, 0),
            (1, 0, 1),
            (6, 0, 6),
            (0, 6, 1),
            (1, 6, 2),
            (6, 6, 0),
        ]
    )
    def test_position(self, index, first_week_day, position):
        i18n.get_locale = flexmock(first_week_day=first_week_day)
        day = WeekDay(index)
        assert day.position == position

    def test_get_name_returns_localized_week_day_name(self):
        day = WeekDay(0)
        assert day.get_name() == u'maanantaina'

    def test_override_get_locale_as_class_method(self):
        day = WeekDay(0)
        assert day.get_name() == u'maanantaina'

    def test_name_delegates_to_get_name(self):
        day = WeekDay(0)
        flexmock(day).should_receive('get_name').and_return(u'maanantaina')
        assert day.name == u'maanantaina'

    def test_unicode(self):
        day = WeekDay(0)
        flexmock(day).should_receive('name').and_return(u'maanantaina')
        assert six.text_type(day) == u'maanantaina'

    def test_str(self):
        day = WeekDay(0)
        flexmock(day).should_receive('name').and_return(u'maanantaina')
        assert str(day) == 'maanantaina'


@pytest.mark.skipif('i18n.babel is None')
class TestWeekDays(object):
    def test_constructor_with_valid_bit_string(self):
        days = WeekDays('1000100')
        assert days._days == set([WeekDay(0), WeekDay(4)])

    @pytest.mark.parametrize(
        'bit_string',
        [
            '000000',    # too short
            '00000000',  # too long
        ]
    )
    def test_constructor_with_bit_string_of_invalid_length(self, bit_string):
        with pytest.raises(ValueError):
            WeekDays(bit_string)

    def test_constructor_with_bit_string_containing_invalid_characters(self):
        with pytest.raises(ValueError):
            WeekDays('foobarz')

    def test_constructor_with_another_week_days_object(self):
        days = WeekDays('0000000')
        another_days = WeekDays(days)
        assert days._days == another_days._days

    def test_representation(self):
        days = WeekDays('0000000')
        assert repr(days) == "WeekDays('0000000')"

    @pytest.mark.parametrize(
        'bit_string',
        [
            '0000000',
            '1000000',
            '0000001',
            '0101000',
            '1111111',
        ]
    )
    def test_as_bit_string(self, bit_string):
        days = WeekDays(bit_string)
        assert days.as_bit_string() == bit_string

    def test_equality_with_equal_week_days_object(self):
        days = WeekDays('0001000')
        days2 = WeekDays('0001000')
        assert days == days2

    def test_equality_with_unequal_week_days_object(self):
        days = WeekDays('0001000')
        days2 = WeekDays('1000000')
        assert days != days2

    def test_equality_with_equal_bit_string(self):
        days = WeekDays('0001000')
        assert days == '0001000'

    def test_equality_with_unequal_bit_string(self):
        days = WeekDays('0001000')
        assert days != '0101000'

    def test_equality_with_unsupported_comparison(self):
        days = WeekDays('0001000')
        assert days != 0

    def test_iterator_starts_from_locales_first_week_day(self):
        i18n.get_locale = lambda: flexmock(first_week_day=1)
        days = WeekDays('1111111')
        indices = list(day.index for day in days)
        assert indices == [1, 2, 3, 4, 5, 6, 0]

    def test_unicode(self):
        i18n.get_locale = lambda: i18n.babel.Locale('fi')
        days = WeekDays('1000100')
        assert six.text_type(days) == u'maanantaina, perjantaina'

    def test_str(self):
        i18n.get_locale = lambda: i18n.babel.Locale('fi')
        days = WeekDays('1000100')
        assert str(days) == 'maanantaina, perjantaina'



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.aggregates import aggregated


@pytest.fixture
def Comment(Base):
    class Comment(Base):
        __tablename__ = 'comment'
        id = sa.Column(sa.Integer, primary_key=True)
        content = sa.Column(sa.Unicode(255))
        thread_id = sa.Column(sa.Integer, sa.ForeignKey('thread.id'))
    return Comment


@pytest.fixture
def Thread(Base, Comment):
    class Thread(Base):
        __tablename__ = 'thread'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @aggregated(
            'comments',
            sa.Column(sa.Integer, default=0)
        )
        def comment_count(self):
            return sa.func.count('1')

        @aggregated('comments', sa.Column(sa.Integer))
        def last_comment_id(self):
            return sa.func.max(Comment.id)

        comments = sa.orm.relationship(
            'Comment',
            backref='thread'
        )

    Thread.last_comment = sa.orm.relationship(
        'Comment',
        primaryjoin='Thread.last_comment_id == Comment.id',
        foreign_keys=[Thread.last_comment_id],
        viewonly=True
    )
    return Thread


@pytest.fixture
def init_models(Comment, Thread):
    pass


class TestAggregateValueGenerationForSimpleModelPaths(object):

    def test_assigns_aggregates_on_insert(self, session, Thread, Comment):
        thread = Thread()
        thread.name = u'some article name'
        session.add(thread)
        comment = Comment(content=u'Some content', thread=thread)
        session.add(comment)
        session.commit()
        session.refresh(thread)
        assert thread.comment_count == 1
        assert thread.last_comment_id == comment.id

    def test_assigns_aggregates_on_separate_insert(
        self,
        session,
        Thread,
        Comment
    ):
        thread = Thread()
        thread.name = u'some article name'
        session.add(thread)
        session.commit()
        comment = Comment(content=u'Some content', thread=thread)
        session.add(comment)
        session.commit()
        session.refresh(thread)
        assert thread.comment_count == 1
        assert thread.last_comment_id == 1

    def test_assigns_aggregates_on_delete(self, session, Thread, Comment):
        thread = Thread()
        thread.name = u'some article name'
        session.add(thread)
        session.commit()
        comment = Comment(content=u'Some content', thread=thread)
        session.add(comment)
        session.commit()
        session.delete(comment)
        session.commit()
        session.refresh(thread)
        assert thread.comment_count == 0
        assert thread.last_comment_id is None



# -*- coding: utf-8 -*-
import pytest
import six

from sqlalchemy_utils import Ltree


class TestLtree(object):
    def test_init(self):
        assert Ltree('path.path') == Ltree(Ltree('path.path'))

    def test_constructor_with_wrong_type(self):
        with pytest.raises(TypeError) as e:
            Ltree(None)
        assert str(e.value) == (
            "Ltree() argument must be a string or an Ltree, not 'NoneType'"
        )

    def test_constructor_with_invalid_code(self):
        with pytest.raises(ValueError) as e:
            Ltree('..')
        assert str(e.value) == "'..' is not a valid ltree path."

    @pytest.mark.parametrize(
        'code',
        (
            'path',
            'path.path',
            '1_.2',
            '_._',
        )
    )
    def test_validate_with_valid_codes(self, code):
        Ltree.validate(code)

    @pytest.mark.parametrize(
        'path',
        (
            '',
            '.',
            'path.',
            'path..path',
            'path.path..path',
            'path.path..',
            'path.äö',
        )
    )
    def test_validate_with_invalid_path(self, path):
        with pytest.raises(ValueError) as e:
            Ltree.validate(path)
        assert str(e.value) == (
            "'{0}' is not a valid ltree path.".format(path)
        )

    @pytest.mark.parametrize(
        ('path', 'length'),
        (
            ('path', 1),
            ('1.1', 2),
            ('1.2.3', 3)
        )
    )
    def test_length(self, path, length):
        assert len(Ltree(path)) == length

    @pytest.mark.parametrize(
        ('path', 'subpath', 'index'),
        (
            ('path.path', 'path', 0),
            ('1.2.3', '2.3', 1),
            ('1.2.3.4', '2.3', 1),
            ('1.2.3.4', '3.4', 2)
        )
    )
    def test_index(self, path, subpath, index):
        assert Ltree(path).index(subpath) == index

    @pytest.mark.parametrize(
        ('path', 'item_slice', 'result'),
        (
            ('path.path', 0, 'path'),
            ('1.1.2.3', slice(1, 3), '1.2'),
            ('1.1.2.3', slice(1, None), '1.2.3'),
        )
    )
    def test_getitem(self, path, item_slice, result):
        assert Ltree(path)[item_slice] == result

    @pytest.mark.parametrize(
        ('path', 'others', 'result'),
        (
            ('1.2.3', ['1.2.3', '1.2'], '1'),
            ('1.2.3.4.5', ['1.2.3', '1.2.3.4'], '1.2'),
            ('1.2.3.4.5', ['3.4', '1.2.3.4'], None),
        )
    )
    def test_lca(self, path, others, result):
        assert Ltree(path).lca(*others) == result

    @pytest.mark.parametrize(
        ('path', 'other', 'result'),
        (
            ('1.2.3', '4.5', '1.2.3.4.5'),
            ('1', '1', '1.1'),
        )
    )
    def test_add(self, path, other, result):
        assert Ltree(path) + other == result

    @pytest.mark.parametrize(
        ('path', 'other', 'result'),
        (
            ('1.2.3', '4.5', '4.5.1.2.3'),
            ('1', '1', '1.1'),
        )
    )
    def test_radd(self, path, other, result):
        assert other + Ltree(path) == result

    @pytest.mark.parametrize(
        ('path', 'other', 'result'),
        (
            ('1.2.3', '4.5', '1.2.3.4.5'),
            ('1', '1', '1.1'),
        )
    )
    def test_iadd(self, path, other, result):
        ltree = Ltree(path)
        ltree += other
        assert ltree == result

    @pytest.mark.parametrize(
        ('path', 'other', 'result'),
        (
            ('1.2.3', '2', True),
            ('1.2.3', '3', True),
            ('1', '1', True),
            ('1', '2', False),
        )
    )
    def test_contains(self, path, other, result):
        assert (other in Ltree(path)) == result

    @pytest.mark.parametrize(
        ('path', 'other', 'result'),
        (
            ('1', '1.2.3', True),
            ('1.2', '1.2.3', True),
            ('1.2.3', '1.2.3', True),
            ('1.2.3', '1', False),
            ('1.2.3', '1.2', False),
            ('1', '1', True),
            ('1', '2', False),
        )
    )
    def test_ancestor_of(self, path, other, result):
        assert Ltree(path).ancestor_of(other) == result

    @pytest.mark.parametrize(
        ('path', 'other', 'result'),
        (
            ('1', '1.2.3', False),
            ('1.2', '1.2.3', False),
            ('1.2', '1.2.3', False),
            ('1.2.3', '1', True),
            ('1.2.3', '1.2', True),
            ('1.2.3', '1.2.3', True),
            ('1', '1', True),
            ('1', '2', False),
        )
    )
    def test_descendant_of(self, path, other, result):
        assert Ltree(path).descendant_of(other) == result

    def test_getitem_with_other_than_slice_or_in(self):
        with pytest.raises(TypeError):
            Ltree('1.2')['something']

    def test_index_raises_value_error_if_subpath_not_found(self):
        with pytest.raises(ValueError):
            Ltree('1.2').index('3')

    def test_equality_operator(self):
        assert Ltree('path.path') == 'path.path'
        assert 'path.path' == Ltree('path.path')
        assert Ltree('path.path') == Ltree('path.path')

    def test_non_equality_operator(self):
        assert Ltree('path.path') != u'path.'
        assert not (Ltree('path.path') != 'path.path')

    def test_hash(self):
        return hash(Ltree('path')) == hash('path')

    def test_repr(self):
        return repr(Ltree('path')) == "Ltree('path')"

    def test_unicode(self):
        ltree = Ltree('path.path')
        assert six.text_type(ltree) == 'path.path'

    def test_str(self):
        ltree = Ltree('path')
        assert str(ltree) == 'path'



import pytest
import sqlalchemy as sa
from flexmock import flexmock

from sqlalchemy_utils import ColorType, types  # noqa


@pytest.fixture
def Document(Base):
    class Document(Base):
        __tablename__ = 'document'
        id = sa.Column(sa.Integer, primary_key=True)
        bg_color = sa.Column(ColorType)

        def __repr__(self):
            return 'Document(%r)' % self.id
    return Document


@pytest.fixture
def init_models(Document):
    pass


@pytest.mark.skipif('types.color.python_colour_type is None')
class TestColorType(object):
    def test_string_parameter_processing(self, session, Document):
        from colour import Color

        flexmock(ColorType).should_receive('_coerce').and_return(
            u'white'
        )
        document = Document(
            bg_color='white'
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.bg_color.hex == Color(u'white').hex

    def test_color_parameter_processing(self, session, Document):
        from colour import Color

        document = Document(bg_color=Color(u'white'))
        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.bg_color.hex == Color(u'white').hex

    def test_scalar_attributes_get_coerced_to_objects(self, Document):
        from colour import Color

        document = Document(bg_color='white')
        assert isinstance(document.bg_color, Color)

    def test_literal_param(self, session, Document):
        clause = Document.bg_color == 'white'
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == "document.bg_color = 'white'"



import pytest
import sqlalchemy as sa
from flexmock import flexmock
from sqlalchemy.dialects.postgresql import HSTORE
from sqlalchemy.orm import aliased

from sqlalchemy_utils import i18n, TranslationHybrid  # noqa


@pytest.fixture
def translation_hybrid():
    return TranslationHybrid('fi', 'en')


@pytest.fixture
def City(Base, translation_hybrid):
    class City(Base):
        __tablename__ = 'city'
        id = sa.Column(sa.Integer, primary_key=True)
        name_translations = sa.Column(HSTORE)
        name = translation_hybrid(name_translations)
        locale = 'en'
    return City


@pytest.fixture
def init_models(City):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
@pytest.mark.skipif('i18n.babel is None')
class TestTranslationHybrid(object):

    def test_using_hybrid_as_constructor(self, City):
        city = City(name='Helsinki')
        assert city.name_translations['fi'] == 'Helsinki'

    def test_if_no_translation_exists_returns_none(self, City):
        city = City()
        assert city.name is None

    def test_custom_default_value(self, City, translation_hybrid):
        translation_hybrid.default_value = 'Some value'
        city = City()
        assert city.name == 'Some value'

    def test_fall_back_to_default_translation(self, City, translation_hybrid):
        city = City(name_translations={'en': 'Helsinki'})
        translation_hybrid.current_locale = 'sv'
        assert city.name == 'Helsinki'

    def test_fallback_to_dynamic_locale(self, City, translation_hybrid):
        translation_hybrid.current_locale = 'en'
        translation_hybrid.default_locale = lambda self: self.locale
        city = City(name_translations={})
        city.locale = 'fi'
        city.name_translations['fi'] = 'Helsinki'

        assert city.name == 'Helsinki'

    def test_fallback_to_attr_dependent_locale(self, City, translation_hybrid):
        translation_hybrid.current_locale = 'en'
        translation_hybrid.default_locale = (
            lambda obj, attr: sorted(getattr(obj, attr).keys())[0]
        )
        city = City(name_translations={})
        city.name_translations['fi'] = 'Helsinki'
        assert city.name == 'Helsinki'
        city.name_translations['de'] = 'Stadt Helsinki'
        assert city.name == 'Stadt Helsinki'

    @pytest.mark.parametrize(
        ('name_translations', 'name'),
        (
            ({'fi': 'Helsinki', 'en': 'Helsing'}, 'Helsinki'),
            ({'en': 'Helsinki'}, 'Helsinki'),
            ({'fi': 'Helsinki'}, 'Helsinki'),
            ({}, None),
        )
    )
    def test_hybrid_as_an_expression(
        self,
        session,
        City,
        name_translations,
        name
    ):
        city = City(name_translations=name_translations)
        session.add(city)
        session.commit()

        assert session.query(City.name).scalar() == name

    def test_dynamic_locale(self, Base):
        translation_hybrid = TranslationHybrid(
            lambda obj: obj.locale,
            'fi'
        )

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            name_translations = sa.Column(HSTORE)
            name = translation_hybrid(name_translations)
            locale = sa.Column(sa.String)

        assert (
            'coalesce(article.name_translations -> article.locale'
            in str(Article.name.expression)
        )

    def test_locales_casted_only_in_compilation_phase(self, Base):
        class LocaleGetter(object):
            def current_locale(self):
                return lambda obj: obj.locale

        flexmock(LocaleGetter).should_receive('current_locale').never()
        translation_hybrid = TranslationHybrid(
            LocaleGetter().current_locale,
            'fi'
        )

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            name_translations = sa.Column(HSTORE)
            name = translation_hybrid(name_translations)
            locale = sa.Column(sa.String)

        Article.name

    def test_no_implicit_join_when_using_aliased_entities(self, session, City):
        # Ensure that queried entities are taken from the alias so that
        # there isn't an extra join to the original entity.
        CityAlias = aliased(City)
        query_str = str(session.query(CityAlias.name))
        assert query_str.endswith('FROM city AS city_1')



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import IntRangeType

intervals = None
inf = -1
try:
    import intervals
    from infinity import inf
except ImportError:
    pass


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column(sa.Integer, primary_key=True)
        persons_at_night = sa.Column(IntRangeType)

        def __repr__(self):
            return 'Building(%r)' % self.id
    return Building


@pytest.fixture
def init_models(Building):
    pass


@pytest.fixture
def create_building(session, Building):
    def create_building(number_range):
        building = Building(
            persons_at_night=number_range
        )

        session.add(building)
        session.commit()
        return session.query(Building).first()
    return create_building


@pytest.mark.skipif('intervals is None')
class NumberRangeTestCase(object):

    def test_nullify_range(self, create_building):
        building = create_building(None)
        assert building.persons_at_night is None

    def test_update_with_none(self, session, create_building):
        interval = intervals.IntInterval([None, None])
        building = create_building(interval)
        building.persons_at_night = None
        assert building.persons_at_night is None
        session.commit()
        assert building.persons_at_night is None

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (0, 4),
        )
    )
    def test_save_number_range(self, create_building, number_range):
        building = create_building(number_range)
        assert building.persons_at_night.lower == 1
        assert building.persons_at_night.upper == 3

    def test_infinite_upper_bound(self, create_building):
        building = create_building([1, inf])
        assert building.persons_at_night.lower == 1
        assert building.persons_at_night.upper == inf

    def test_infinite_lower_bound(self, create_building):
        building = create_building([-inf, 1])
        assert building.persons_at_night.lower == -inf
        assert building.persons_at_night.upper == 1

    def test_nullify_number_range(self, session, Building):
        building = Building(
            persons_at_night=intervals.IntInterval([1, 3])
        )

        session.add(building)
        session.commit()

        building = session.query(Building).first()
        building.persons_at_night = None
        session.commit()

        building = session.query(Building).first()
        assert building.persons_at_night is None

    def test_integer_coercion(self, Building):
        building = Building(persons_at_night=15)
        assert building.persons_at_night.lower == 15
        assert building.persons_at_night.upper == 15


@pytest.mark.usefixtures('postgresql_dsn')
class TestIntRangeTypeOnPostgres(NumberRangeTestCase):
    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (0, 4)
        )
    )
    def test_eq_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night == number_range)
        )
        assert query.count()

    @pytest.mark.parametrize(
        ('number_range', 'length'),
        (
            ([1, 3], 2),
            ([1, 1], 0),
            ([-1, 1], 2),
            ([-inf, 1], None),
            ([0, inf], None),
            ([0, 0], 0),
            ([-3, -1], 2)
        )
    )
    def test_length(
        self,
        session,
        Building,
        create_building,
        number_range,
        length
    ):
        create_building(number_range)
        query = (
            session.query(Building.persons_at_night.length)
        )
        assert query.scalar() == length

    @pytest.mark.parametrize(
        'number_range',
        (
            [[1, 3]],
            [(0, 4)],
        )
    )
    def test_in_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night.in_(number_range))
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (0, 4),
        )
    )
    def test_rshift_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([5, 6])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night >> number_range)
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (0, 4),
        )
    )
    def test_lshift_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([-1, 0])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night << number_range)
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (1, 3),
            2
        )
    )
    def test_contains_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night.contains(number_range))
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (0, 8),
            (-inf, inf)
        )
    )
    def test_contained_by_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night.contained_by(number_range))
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [2, 5],
            0
        )
    )
    def test_not_in_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(~ Building.persons_at_night.in_([number_range]))
        )
        assert query.count()

    def test_eq_with_query_arg(self, session, Building, create_building):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(
                Building.persons_at_night ==
                session.query(Building.persons_at_night)
            ).order_by(Building.persons_at_night).limit(1)
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 2],
            (0, 4),
            [0, 3],
            0,
            1,
        )
    )
    def test_ge_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night >= number_range)
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [0, 2],
            0,
            [-inf, 2]
        )
    )
    def test_gt_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night > number_range)
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 4],
            4,
            [2, inf]
        )
    )
    def test_le_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night <= number_range)
        )
        assert query.count()

    @pytest.mark.parametrize(
        'number_range',
        (
            [2, 4],
            4,
            [1, inf]
        )
    )
    def test_lt_operator(
        self,
        session,
        Building,
        create_building,
        number_range
    ):
        create_building([1, 3])
        query = (
            session.query(Building)
            .filter(Building.persons_at_night < number_range)
        )
        assert query.count()

    def test_literal_param(self, session, Building):
        clause = Building.persons_at_night == [1, 3]
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == "building.persons_at_night = '[1, 3]'"


class TestNumberRangeTypeOnSqlite(NumberRangeTestCase):
    pass



import operator

import pytest
import six

from sqlalchemy_utils import Country, i18n


@pytest.fixture
def set_get_locale():
    i18n.get_locale = lambda: i18n.babel.Locale('en')


@pytest.mark.skipif('i18n.babel is None')
@pytest.mark.usefixtures('set_get_locale')
class TestCountry(object):

    def test_init(self):
        assert Country(u'FI') == Country(Country(u'FI'))

    def test_constructor_with_wrong_type(self):
        with pytest.raises(TypeError) as e:
            Country(None)
        assert str(e.value) == (
            "Country() argument must be a string or a country, not 'NoneType'"
        )

    def test_constructor_with_invalid_code(self):
        with pytest.raises(ValueError) as e:
            Country('SomeUnknownCode')
        assert str(e.value) == (
            'Could not convert string to country code: SomeUnknownCode'
        )

    @pytest.mark.parametrize(
        'code',
        (
            'FI',
            'US',
        )
    )
    def test_validate_with_valid_codes(self, code):
        Country.validate(code)

    def test_validate_with_invalid_code(self):
        with pytest.raises(ValueError) as e:
            Country.validate('SomeUnknownCode')
        assert str(e.value) == (
            'Could not convert string to country code: SomeUnknownCode'
        )

    def test_equality_operator(self):
        assert Country(u'FI') == u'FI'
        assert u'FI' == Country(u'FI')
        assert Country(u'FI') == Country(u'FI')

    def test_non_equality_operator(self):
        assert Country(u'FI') != u'sv'
        assert not (Country(u'FI') != u'FI')

    @pytest.mark.parametrize(
        'op, code_left, code_right, is_',
        [
            (operator.lt, u'ES', u'FI', True),
            (operator.lt, u'FI', u'ES', False),
            (operator.lt, u'ES', u'ES', False),

            (operator.le, u'ES', u'FI', True),
            (operator.le, u'FI', u'ES', False),
            (operator.le, u'ES', u'ES', True),

            (operator.ge, u'ES', u'FI', False),
            (operator.ge, u'FI', u'ES', True),
            (operator.ge, u'ES', u'ES', True),

            (operator.gt, u'ES', u'FI', False),
            (operator.gt, u'FI', u'ES', True),
            (operator.gt, u'ES', u'ES', False),
        ]
    )
    def test_ordering(self, op, code_left, code_right, is_):
        country_left = Country(code_left)
        country_right = Country(code_right)
        assert op(country_left, country_right) is is_
        assert op(country_left, code_right) is is_
        assert op(code_left, country_right) is is_

    def test_hash(self):
        return hash(Country('FI')) == hash('FI')

    def test_repr(self):
        return repr(Country('FI')) == "Country('FI')"

    def test_unicode(self):
        country = Country('FI')
        assert six.text_type(country) == u'Finland'

    def test_str(self):
        country = Country('FI')
        assert str(country) == 'Finland'



from decimal import Decimal

import pytest
import sqlalchemy as sa

from sqlalchemy_utils.aggregates import aggregated


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        price = sa.Column(sa.Numeric)

        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))
    return Product


@pytest.fixture
def Catalog(Base, Product):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @aggregated('products', sa.Column(sa.Numeric, default=0))
        def net_worth(self):
            return sa.func.sum(Product.price)

        products = sa.orm.relationship('Product', backref='catalog')
    return Catalog


@pytest.fixture
def init_models(Product, Catalog):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestLazyEvaluatedSelectExpressionsForAggregates(object):

    def test_assigns_aggregates_on_insert(self, session, Product, Catalog):
        catalog = Catalog(
            name=u'Some catalog'
        )
        session.add(catalog)
        session.commit()
        product = Product(
            name=u'Some product',
            price=Decimal('1000'),
            catalog=catalog
        )
        session.add(product)
        session.commit()
        session.refresh(catalog)
        assert catalog.net_worth == Decimal('1000')

    def test_assigns_aggregates_on_update(self, session, Product, Catalog):
        catalog = Catalog(
            name=u'Some catalog'
        )
        session.add(catalog)
        session.commit()
        product = Product(
            name=u'Some product',
            price=Decimal('1000'),
            catalog=catalog
        )
        session.add(product)
        session.commit()
        product.price = Decimal('500')
        session.commit()
        session.refresh(catalog)
        assert catalog.net_worth == Decimal('500')



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import Country, CountryType, i18n  # noqa


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        country = sa.Column(CountryType)

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.mark.skipif('i18n.babel is None')
class TestCountryType(object):

    def test_parameter_processing(self, session, User):
        user = User(
            country=Country(u'FI')
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.country.name == u'Finland'

    def test_scalar_attributes_get_coerced_to_objects(self, User):
        user = User(country='FI')
        assert isinstance(user.country, Country)

    def test_literal_param(self, session, User):
        clause = User.country == 'FI'
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == '"user".country = \'FI\''



"""
SQLAlchemy-Utils
----------------

Various utility functions and custom data types for SQLAlchemy.
"""
from setuptools import setup, find_packages
import os
import re
import sys


HERE = os.path.dirname(os.path.abspath(__file__))
PY3 = sys.version_info[0] == 3


def get_version():
    filename = os.path.join(HERE, 'sqlalchemy_utils', '__init__.py')
    with open(filename) as f:
        contents = f.read()
    pattern = r"^__version__ = '(.*?)'$"
    return re.search(pattern, contents, re.MULTILINE).group(1)


extras_require = {
    'test': [
        'pytest>=2.7.1',
        'Pygments>=1.2',
        'Jinja2>=2.3',
        'docutils>=0.10',
        'flexmock>=0.9.7',
        'mock==2.0.0',
        'psycopg2>=2.5.1',
        'pg8000>=1.12.4',
        'pytz>=2014.2',
        'python-dateutil>=2.6',
        'pymysql',
        'flake8>=2.4.0',
        'isort>=4.2.2',
        'pyodbc',
    ],
    'anyjson': ['anyjson>=0.3.3'],
    'babel': ['Babel>=1.3'],
    'arrow': ['arrow>=0.3.4'],
    'intervals': ['intervals>=0.7.1'],
    'phone': ['phonenumbers>=5.9.2'],
    'password': ['passlib >= 1.6, < 2.0'],
    'color': ['colour>=0.0.4'],
    'ipaddress': ['ipaddr'] if not PY3 else [],
    'enum': ['enum34'] if sys.version_info < (3, 4) else [],
    'timezone': ['python-dateutil'],
    'url': ['furl >= 0.4.1'],
    'encrypted': ['cryptography>=0.6']
}


# Add all optional dependencies to testing requirements.
test_all = []
for name, requirements in sorted(extras_require.items()):
    test_all += requirements
extras_require['test_all'] = test_all


setup(
    name='SQLAlchemy-Utils',
    version=get_version(),
    url='https://github.com/kvesteri/sqlalchemy-utils',
    license='BSD',
    author='Konsta Vesterinen, Ryan Leckey, Janne Vanhala, Vesa Uimonen',
    author_email='konsta@fastmonkeys.com',
    description=(
        'Various utility functions for SQLAlchemy.'
    ),
    long_description=__doc__,
    packages=find_packages('.', exclude=['tests', 'tests.*']),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'six',
        'SQLAlchemy>=1.0'
    ],
    extras_require=extras_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import dependent_objects, get_referencing_foreign_keys


class TestDependentObjects(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            first_name = sa.Column(sa.Unicode(255))
            last_name = sa.Column(sa.Unicode(255))
        return User

    @pytest.fixture
    def Article(self, Base, User):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_id = sa.Column(sa.Integer, sa.ForeignKey('user.id'))
            owner_id = sa.Column(
                sa.Integer, sa.ForeignKey('user.id', ondelete='SET NULL')
            )

            author = sa.orm.relationship(User, foreign_keys=[author_id])
            owner = sa.orm.relationship(User, foreign_keys=[owner_id])
        return Article

    @pytest.fixture
    def BlogPost(self, Base, User):
        class BlogPost(Base):
            __tablename__ = 'blog_post'
            id = sa.Column(sa.Integer, primary_key=True)
            owner_id = sa.Column(
                sa.Integer, sa.ForeignKey('user.id', ondelete='CASCADE')
            )

            owner = sa.orm.relationship(User)
        return BlogPost

    @pytest.fixture
    def init_models(self, User, Article, BlogPost):
        pass

    def test_returns_all_dependent_objects(self, session, User, Article):
        user = User(first_name=u'John')
        articles = [
            Article(author=user),
            Article(),
            Article(owner=user),
            Article(author=user, owner=user)
        ]
        session.add_all(articles)
        session.commit()

        deps = list(dependent_objects(user))
        assert len(deps) == 3
        assert articles[0] in deps
        assert articles[2] in deps
        assert articles[3] in deps

    def test_with_foreign_keys_parameter(
        self,
        session,
        User,
        Article,
        BlogPost
    ):
        user = User(first_name=u'John')
        objects = [
            Article(author=user),
            Article(),
            Article(owner=user),
            Article(author=user, owner=user),
            BlogPost(owner=user)
        ]
        session.add_all(objects)
        session.commit()

        deps = list(
            dependent_objects(
                user,
                (
                    fk for fk in get_referencing_foreign_keys(User)
                    if fk.ondelete == 'RESTRICT' or fk.ondelete is None
                )
            ).limit(5)
        )
        assert len(deps) == 2
        assert objects[0] in deps
        assert objects[3] in deps


class TestDependentObjectsWithColumnAliases(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            first_name = sa.Column(sa.Unicode(255))
            last_name = sa.Column(sa.Unicode(255))
        return User

    @pytest.fixture
    def Article(self, Base, User):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_id = sa.Column(
                '_author_id', sa.Integer, sa.ForeignKey('user.id')
            )
            owner_id = sa.Column(
                '_owner_id',
                sa.Integer, sa.ForeignKey('user.id', ondelete='SET NULL')
            )

            author = sa.orm.relationship(User, foreign_keys=[author_id])
            owner = sa.orm.relationship(User, foreign_keys=[owner_id])
        return Article

    @pytest.fixture
    def BlogPost(self, Base, User):
        class BlogPost(Base):
            __tablename__ = 'blog_post'
            id = sa.Column(sa.Integer, primary_key=True)
            owner_id = sa.Column(
                '_owner_id',
                sa.Integer, sa.ForeignKey('user.id', ondelete='CASCADE')
            )

            owner = sa.orm.relationship(User)
        return BlogPost

    @pytest.fixture
    def init_models(self, User, Article, BlogPost):
        pass

    def test_returns_all_dependent_objects(self, session, User, Article):
        user = User(first_name=u'John')
        articles = [
            Article(author=user),
            Article(),
            Article(owner=user),
            Article(author=user, owner=user)
        ]
        session.add_all(articles)
        session.commit()

        deps = list(dependent_objects(user))
        assert len(deps) == 3
        assert articles[0] in deps
        assert articles[2] in deps
        assert articles[3] in deps

    def test_with_foreign_keys_parameter(
        self,
        session,
        User,
        Article,
        BlogPost
    ):
        user = User(first_name=u'John')
        objects = [
            Article(author=user),
            Article(),
            Article(owner=user),
            Article(author=user, owner=user),
            BlogPost(owner=user)
        ]
        session.add_all(objects)
        session.commit()

        deps = list(
            dependent_objects(
                user,
                (
                    fk for fk in get_referencing_foreign_keys(User)
                    if fk.ondelete == 'RESTRICT' or fk.ondelete is None
                )
            ).limit(5)
        )
        assert len(deps) == 2
        assert objects[0] in deps
        assert objects[3] in deps


class TestDependentObjectsWithManyReferences(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            first_name = sa.Column(sa.Unicode(255))
            last_name = sa.Column(sa.Unicode(255))
        return User

    @pytest.fixture
    def BlogPost(self, Base, User):
        class BlogPost(Base):
            __tablename__ = 'blog_post'
            id = sa.Column(sa.Integer, primary_key=True)
            author_id = sa.Column(sa.Integer, sa.ForeignKey('user.id'))
            author = sa.orm.relationship(User)
        return BlogPost

    @pytest.fixture
    def Article(self, Base, User):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_id = sa.Column(sa.Integer, sa.ForeignKey('user.id'))
            author = sa.orm.relationship(User)
        return Article

    @pytest.fixture
    def init_models(self, User, BlogPost, Article):
        pass

    def test_with_many_dependencies(self, session, User, Article, BlogPost):
        user = User(first_name=u'John')
        objects = [
            Article(author=user),
            BlogPost(author=user)
        ]
        session.add_all(objects)
        session.commit()
        deps = list(dependent_objects(user))
        assert len(deps) == 2


class TestDependentObjectsWithCompositeKeys(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)
        return User

    @pytest.fixture
    def Article(self, Base, User):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
            )

            author = sa.orm.relationship(User)
        return Article

    @pytest.fixture
    def init_models(self, User, Article):
        pass

    def test_returns_all_dependent_objects(self, session, User, Article):
        user = User(first_name=u'John', last_name=u'Smith')
        articles = [
            Article(author=user),
            Article(),
            Article(),
            Article(author=user)
        ]
        session.add_all(articles)
        session.commit()

        deps = list(dependent_objects(user))
        assert len(deps) == 2
        assert articles[0] in deps
        assert articles[3] in deps


class TestDependentObjectsWithSingleTableInheritance(object):

    @pytest.fixture
    def Category(self, Base):
        class Category(Base):
            __tablename__ = 'category'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return Category

    @pytest.fixture
    def TextItem(self, Base, Category):
        class TextItem(Base):
            __tablename__ = 'text_item'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
            category_id = sa.Column(
                sa.Integer,
                sa.ForeignKey(Category.id)
            )
            category = sa.orm.relationship(
                Category,
                backref=sa.orm.backref(
                    'articles'
                )
            )
            type = sa.Column(sa.Unicode(255))

            __mapper_args__ = {
                'polymorphic_on': type,
            }
        return TextItem

    @pytest.fixture
    def Article(self, TextItem):
        class Article(TextItem):
            __mapper_args__ = {
                'polymorphic_identity': u'article'
            }
        return Article

    @pytest.fixture
    def BlogPost(self, TextItem):
        class BlogPost(TextItem):
            __mapper_args__ = {
                'polymorphic_identity': u'blog_post'
            }
        return BlogPost

    @pytest.fixture
    def init_models(self, Category, TextItem, Article, BlogPost):
        pass

    def test_returns_all_dependent_objects(self, session, Category, Article):
        category1 = Category(name=u'Category #1')
        category2 = Category(name=u'Category #2')
        articles = [
            Article(category=category1),
            Article(category=category1),
            Article(category=category2),
            Article(category=category2),
        ]
        session.add_all(articles)
        session.commit()

        deps = list(dependent_objects(category1))
        assert len(deps) == 2
        assert articles[0] in deps
        assert articles[1] in deps



import pytest
import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy_utils import get_hybrid_properties


@pytest.fixture
def Category(Base):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @hybrid_property
        def lowercase_name(self):
            return self.name.lower()

        @lowercase_name.expression
        def lowercase_name(cls):
            return sa.func.lower(cls.name)
    return Category


class TestGetHybridProperties(object):

    def test_declarative_model(self, Category):
        assert (
            list(get_hybrid_properties(Category).keys()) ==
            ['lowercase_name']
        )

    def test_mapper(self, Category):
        assert (
            list(get_hybrid_properties(sa.inspect(Category)).keys()) ==
            ['lowercase_name']
        )

    def test_aliased_class(self, Category):
        props = get_hybrid_properties(sa.orm.aliased(Category))
        assert list(props.keys()) == ['lowercase_name']



import pytest
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from sqlalchemy_utils import Asterisk, row_to_json


@pytest.fixture
def assert_startswith(session):
    def assert_startswith(query, query_part):
        assert str(
            query.compile(dialect=postgresql.dialect())
        ).startswith(query_part)
        # Check that query executes properly
        session.execute(query)
    return assert_startswith


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        content = sa.Column(sa.UnicodeText)
    return Article


class TestAsterisk(object):
    def test_with_table_object(self):
        Base = sa.ext.declarative.declarative_base()

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)

        assert str(Asterisk(Article.__table__)) == 'article.*'

    def test_with_quoted_identifier(self):
        Base = sa.ext.declarative.declarative_base()

        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)

        assert str(Asterisk(User.__table__).compile(
            dialect=postgresql.dialect()
        )) == '"user".*'


class TestRowToJson(object):
    def test_compiler_with_default_dialect(self):
        assert str(row_to_json(sa.text('article.*'))) == (
            'row_to_json(article.*)'
        )

    def test_compiler_with_postgresql(self):
        assert str(row_to_json(sa.text('article.*')).compile(
            dialect=postgresql.dialect()
        )) == 'row_to_json(article.*)'

    def test_type(self):
        assert isinstance(
            sa.func.row_to_json(sa.text('article.*')).type,
            postgresql.JSON
        )



from datetime import date, datetime, time

import pytest
import sqlalchemy as sa

from sqlalchemy_utils import ColorType, EncryptedType, PhoneNumberType
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesEngine,
    AesGcmEngine,
    DatetimeHandler,
    FernetEngine,
    InvalidCiphertextError
)

cryptography = None
try:
    import cryptography  # noqa
except ImportError:
    pass


@pytest.fixture
def User(Base, encryption_engine, test_key, padding_mechanism):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)

        username = sa.Column(EncryptedType(
            sa.Unicode,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        access_token = sa.Column(EncryptedType(
            sa.String,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        is_active = sa.Column(EncryptedType(
            sa.Boolean,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        accounts_num = sa.Column(EncryptedType(
            sa.Integer,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        phone = sa.Column(EncryptedType(
            PhoneNumberType,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        color = sa.Column(EncryptedType(
            ColorType,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        date = sa.Column(EncryptedType(
            sa.Date,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        time = sa.Column(EncryptedType(
            sa.Time,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        datetime = sa.Column(EncryptedType(
            sa.DateTime,
            test_key,
            encryption_engine,
            padding_mechanism)
        )

        enum = sa.Column(EncryptedType(
            sa.Enum('One', name='user_enum_t'),
            test_key,
            encryption_engine,
            padding_mechanism)
        )

    return User


@pytest.fixture
def test_key():
    return 'secretkey1234'


@pytest.fixture
def user_name():
    return u'someone'


@pytest.fixture
def user_phone():
    return u'(555) 555-5555'


@pytest.fixture
def user_color():
    return u'#fff'


@pytest.fixture
def user_enum():
    return 'One'


@pytest.fixture
def user_date():
    return date(2010, 10, 2)


@pytest.fixture
def user_time():
    return time(10, 12)


@pytest.fixture
def user_datetime():
    return datetime(2010, 10, 2, 10, 12, 45, 2334)


@pytest.fixture
def test_token():
    import string
    import random
    token = ''
    characters = string.ascii_letters + string.digits
    for i in range(60):
        token += ''.join(random.choice(characters))
    return token


@pytest.fixture
def active():
    return True


@pytest.fixture
def accounts_num():
    return 2


@pytest.fixture
def user(
    request,
    session,
    User,
    user_name,
    user_phone,
    user_color,
    user_date,
    user_time,
    user_enum,
    user_datetime,
    test_token,
    active,
    accounts_num
):
    # set the values to the user object
    user = User()
    user.username = user_name
    user.phone = user_phone
    user.color = user_color
    user.date = user_date
    user.time = user_time
    user.enum = user_enum
    user.datetime = user_datetime
    user.access_token = test_token
    user.is_active = active
    user.accounts_num = accounts_num
    session.add(user)
    session.commit()

    return session.query(User).get(user.id)


@pytest.fixture
def datetime_with_micro_and_timezone():
    import pytz
    tz = pytz.timezone('Pacific/Tahiti')
    return datetime(2017, 8, 21, 4, 26, 36, 523010, tzinfo=tz)


@pytest.fixture
def datetime_with_micro():
    return datetime(2017, 8, 21, 10, 12, 45, 22)


@pytest.fixture
def datetime_simple():
    return datetime(2017, 8, 21, 10, 12, 45)


@pytest.fixture
def time_with_micro():
    return time(10, 12, 45, 22)


@pytest.fixture
def time_simple():
    return time(10, 12, 45)


@pytest.fixture
def date_simple():
    return date(2017, 8, 21)


@pytest.mark.skipif('cryptography is None')
class EncryptedTypeTestCase(object):

    @pytest.fixture
    def Team(self, Base, encryption_engine, padding_mechanism):
        self._team_key = None

        class Team(Base):
            __tablename__ = 'team'
            id = sa.Column(sa.Integer, primary_key=True)
            key = sa.Column(sa.String(50))
            name = sa.Column(EncryptedType(
                sa.Unicode,
                lambda: self._team_key,
                encryption_engine,
                padding_mechanism)
            )
        return Team

    @pytest.fixture
    def init_models(self, User, Team):
        pass

    def test_unicode(self, user, user_name):
        assert user.username == user_name

    def test_string(self, user, test_token):
        assert user.access_token == test_token

    def test_boolean(self, user, active):
        assert user.is_active == active

    def test_integer(self, user, accounts_num):
        assert user.accounts_num == accounts_num

    def test_phone_number(self, user, user_phone):
        assert str(user.phone) == user_phone

    def test_color(self, user, user_color):
        assert user.color.hex == user_color

    def test_date(self, user, user_date):
        assert user.date == user_date

    def test_datetime(self, user, user_datetime):
        assert user.datetime == user_datetime

    def test_time(self, user, user_time):
        assert user.time == user_time

    def test_enum(self, user, user_enum):
        assert user.enum == user_enum

    def test_lookup_key(self, session, Team):
        # Add teams
        self._team_key = 'one'
        team = Team(key=self._team_key, name=u'One')
        session.add(team)
        session.commit()
        team_1_id = team.id

        self._team_key = 'two'
        team = Team(key=self._team_key)
        team.name = u'Two'
        session.add(team)
        session.commit()
        team_2_id = team.id

        # Lookup teams
        self._team_key = session.query(Team.key).filter_by(
            id=team_1_id
        ).one()[0]

        team = session.query(Team).get(team_1_id)

        assert team.name == u'One'

        session.expunge_all()

        self._team_key = session.query(Team.key).filter_by(
            id=team_2_id
        ).one()[0]

        team = session.query(Team).get(team_2_id)

        assert team.name == u'Two'

        session.expunge_all()

        # Remove teams
        session.query(Team).delete()
        session.commit()


class AesEncryptedTypeTestCase(EncryptedTypeTestCase):

    @pytest.fixture
    def encryption_engine(self):
        return AesEngine

    def test_lookup_by_encrypted_string(self, session, User, user, user_name):
        test = session.query(User).filter(
            User.username == user_name
        ).first()

        assert test.username == user.username


class TestAesEncryptedTypeWithPKCS5Padding(AesEncryptedTypeTestCase):

    @pytest.fixture
    def padding_mechanism(self):
        return 'pkcs5'


class TestAesEncryptedTypeWithOneAndZeroesPadding(AesEncryptedTypeTestCase):

    @pytest.fixture
    def padding_mechanism(self):
        return 'oneandzeroes'


class TestAesEncryptedTypeWithZeroesPadding(AesEncryptedTypeTestCase):

    @pytest.fixture
    def padding_mechanism(self):
        return 'zeroes'


class TestAesEncryptedTypeTestcaseWithNaivePadding(AesEncryptedTypeTestCase):

    @pytest.fixture
    def padding_mechanism(self):
        return 'naive'

    def test_decrypt_raises_value_error_with_invalid_key(self, session, Team):
        self._team_key = 'one'
        team = Team(key=self._team_key, name=u'One')
        session.add(team)
        session.commit()

        self._team_key = 'notone'
        with pytest.raises(ValueError):
            assert team.name == u'One'


class TestFernetEncryptedTypeTestCase(EncryptedTypeTestCase):

    @pytest.fixture
    def encryption_engine(self):
        return FernetEngine

    @pytest.fixture
    def padding_mechanism(self):
        return None


class TestDatetimeHandler(object):

    def test_datetime_with_micro_and_timezone(
        self, datetime_with_micro_and_timezone
    ):
        original_datetime = datetime_with_micro_and_timezone
        original_datetime_isoformat = original_datetime.isoformat()
        python_type = datetime

        assert DatetimeHandler.process_value(
            original_datetime_isoformat,
            python_type
        ) == original_datetime

    def test_datetime_with_micro(self, datetime_with_micro):
        original_datetime = datetime_with_micro
        original_datetime_isoformat = original_datetime.isoformat()
        python_type = datetime

        assert DatetimeHandler.process_value(
            original_datetime_isoformat,
            python_type
        ) == original_datetime

    def test_datetime_simple(self, datetime_simple):
        original_datetime = datetime_simple
        original_datetime_isoformat = original_datetime.isoformat()
        python_type = datetime

        assert DatetimeHandler.process_value(
            original_datetime_isoformat,
            python_type
        ) == original_datetime

    def test_time_with_micro(self, time_with_micro):
        original_time = time_with_micro
        original_time_isoformat = original_time.isoformat()
        python_type = time

        assert DatetimeHandler.process_value(
            original_time_isoformat,
            python_type
        ) == original_time

    def test_time_simple(self, time_simple):
        original_time = time_simple
        original_time_isoformat = original_time.isoformat()
        python_type = time

        assert DatetimeHandler.process_value(
            original_time_isoformat,
            python_type
        ) == original_time

    def test_date_simple(self, date_simple):
        original_date = date_simple
        original_date_isoformat = original_date.isoformat()
        python_type = date

        assert DatetimeHandler.process_value(
            original_date_isoformat,
            python_type
        ) == original_date


@pytest.mark.skipif('cryptography is None')
class TestAesGcmEngine(object):
    KEY = b'0123456789ABCDEF'

    def setup_method(self):
        self.engine = AesGcmEngine()
        self.engine._initialize_engine(TestAesGcmEngine.KEY)

    def test_roundtrip(self):
        for l in range(0, 36):
            plaintext = '0123456789abcdefghijklmnopqrstuvwxyz'[:l]
            encrypted = self.engine.encrypt(plaintext)
            decrypted = self.engine.decrypt(encrypted)
            assert plaintext == decrypted, "Round-trip failed for len: %d" % l

    def test_modified_iv_fails_to_decrypt(self):
        plaintext = 'abcdefgh'
        encrypted = self.engine.encrypt(plaintext)
        # 3rd char will be IV. Modify it
        POS = 3
        encrypted = encrypted[:POS] + \
            (b'A' if encrypted[POS] != b'A' else b'B') + \
            encrypted[POS + 1:]
        with pytest.raises(InvalidCiphertextError):
            self.engine.decrypt(encrypted)

    def test_modified_tag_fails_to_decrypt(self):
        plaintext = 'abcdefgh'
        encrypted = self.engine.encrypt(plaintext)
        # 19th char will be tag. Modify it
        POS = 19
        encrypted = encrypted[:POS] + \
            (b'A' if encrypted[POS] != b'A' else b'B') + \
            encrypted[POS + 1:]
        with pytest.raises(InvalidCiphertextError):
            self.engine.decrypt(encrypted)

    def test_modified_ciphertext_fails_to_decrypt(self):
        plaintext = 'abcdefgh'
        encrypted = self.engine.encrypt(plaintext)
        # 43rd char will be ciphertext. Modify it
        POS = 43
        encrypted = encrypted[:POS] + \
            (b'A' if encrypted[POS] != b'A' else b'B') + \
            encrypted[POS + 1:]
        with pytest.raises(InvalidCiphertextError):
            self.engine.decrypt(encrypted)

    def test_too_short_ciphertext_fails_to_decrypt(self):
        plaintext = 'abcdefgh'
        encrypted = self.engine.encrypt(plaintext)[:20]
        with pytest.raises(InvalidCiphertextError):
            self.engine.decrypt(encrypted)

    def test_different_ciphertexts_each_time(self):
        plaintext = 'abcdefgh'
        encrypted1 = self.engine.encrypt(plaintext)
        encrypted2 = self.engine.encrypt(plaintext)
        assert self.engine.decrypt(encrypted1) == \
            self.engine.decrypt(encrypted2)
        # The following has a very low probability of failing
        # accidentally (2^-96)
        assert encrypted1 != encrypted2


class TestAesGcmEncryptedType(EncryptedTypeTestCase):

    @pytest.fixture
    def encryption_engine(self):
        return AesGcmEngine

    # GCM doesn't need padding. This is here just because we're reusing test
    # code that requires this
    @pytest.fixture
    def padding_mechanism(self):
        return 'pkcs5'

    def test_lookup_by_encrypted_string(self, session, User, user, user_name):
        test = session.query(User).filter(
            User.username == "someonex"
        ).first()

        # With probability 1-2^-96, the 2 different encryptions will choose a
        # different IV, and will therefore result in different ciphertexts.
        # Thus, the 2 values will almost certainly be different, even though
        # we're really searching for the same username. Hence, the above search
        # will fail
        assert test is None
